import json
import logging
import random
import warnings
from collections import namedtuple

import ujson
from pyramid.settings import asbool

from kinto.core.decorators import deprecate_kwargs

from . import generators


class Missing:
    """Dummy value to represent a value that is completely absent from an object.

    Handling these correctly is important for pagination.
    """

    pass


MISSING = Missing()


logger = logging.getLogger(__name__)


Filter = namedtuple("Filter", ["field", "value", "operator"])
"""Filtering properties."""

Sort = namedtuple("Sort", ["field", "direction"])
"""Sorting properties."""

DEFAULT_ID_FIELD = "id"
DEFAULT_MODIFIED_FIELD = "last_modified"
DEFAULT_DELETED_FIELD = "deleted"

_HEARTBEAT_DELETE_RATE = 0.6
_HEARTBEAT_RESOURCE_NAME = "__heartbeat__"
_HEART_PARENT_ID = _HEARTBEAT_RESOURCE_NAME
_HEARTBEAT_OBJECT = {"__heartbeat__": True}


class StorageBase:
    """Storage abstraction used by resource views.

    It is meant to be instantiated at application startup.
    Any operation may raise a `HTTPServiceUnavailable` error if an error
    occurs with the underlying service.

    Configuration can be changed to choose which storage backend will
    persist the objects.

    :raises: :exc:`~pyramid:pyramid.httpexceptions.HTTPServiceUnavailable`
    """

    id_generator = generators.UUID4()
    """Id generator used when no one is provided for create."""

    def __init__(self, strict_json=True):
        """initialize json (de)serializer to be the strict, slow json or ujson"""
        if strict_json:
            self.json = json
        else:
            self.json = ujson

    def initialize_schema(self, dry_run=False):
        """Create every necessary objects (like tables or indices) in the
        backend.

        This is executed when the ``kinto migrate`` command is run.

        :param bool dry_run: simulate instead of executing the operations.
        """
        raise NotImplementedError

    def flush(self):
        """Remove **every** object from this storage.
        """
        raise NotImplementedError

    def resource_timestamp(self, resource_name, parent_id):
        """Get the highest timestamp of every objects in this `resource_name` for
        this `parent_id`.

        .. note::

            This should take deleted objects into account.

        :param str resource_name: the resource name.
        :param str parent_id: the resource parent.

        :returns: the latest timestamp of the resource.
        :rtype: int
        """
        raise NotImplementedError

    def create(
        self,
        resource_name,
        parent_id,
        obj,
        id_generator=None,
        id_field=DEFAULT_ID_FIELD,
        modified_field=DEFAULT_MODIFIED_FIELD,
    ):
        """Create the specified `obj` in this `resource_name` for this `parent_id`.
        Assign the id to the object, using the attribute
        :attr:`kinto.core.resource.model.Model.id_field`.

        .. note::

            This will update the resource timestamp.

        :raises: :exc:`kinto.core.storage.exceptions.UnicityError`

        :param str resource_name: the resource name.
        :param str parent_id: the resource parent.
        :param dict obj: the object to create.

        :returns: the newly created object.
        :rtype: dict
        """
        raise NotImplementedError

    def get(
        self,
        resource_name,
        parent_id,
        object_id,
        id_field=DEFAULT_ID_FIELD,
        modified_field=DEFAULT_MODIFIED_FIELD,
    ):
        """Retrieve the object with specified `object_id`, or raise error
        if not found.

        :raises: :exc:`kinto.core.storage.exceptions.ObjectNotFoundError`

        :param str resource_name: the resource name.
        :param str parent_id: the resource parent.

        :param str object_id: unique identifier of the object

        :returns: the stored object.
        :rtype: dict
        """
        raise NotImplementedError

    def update(
        self,
        resource_name,
        parent_id,
        object_id,
        obj,
        id_field=DEFAULT_ID_FIELD,
        modified_field=DEFAULT_MODIFIED_FIELD,
    ):
        """Overwrite the `obj` with the specified `object_id`.

        If the specified id is not found, the object is created with the
        specified id.

        .. note::

            This will update the resource timestamp.

        :param str resource_name: the resource name.
        :param str parent_id: the resource parent.
        :param str object_id: unique identifier of the object
        :param dict obj: the object to update or create.

        :returns: the updated object.
        :rtype: dict
        """
        raise NotImplementedError

    def delete(
        self,
        resource_name,
        parent_id,
        object_id,
        id_field=DEFAULT_ID_FIELD,
        with_deleted=True,
        modified_field=DEFAULT_MODIFIED_FIELD,
        deleted_field=DEFAULT_DELETED_FIELD,
        last_modified=None,
    ):
        """Delete the object with specified `object_id`, and raise error
        if not found.

        Deleted objects must be removed from the database, but their ids and
        timestamps of deletion must be tracked for synchronization purposes.
        (See :meth:`kinto.core.storage.StorageBase.get_all`)

        .. note::

            This will update the resource timestamp.

        :raises: :exc:`kinto.core.storage.exceptions.ObjectNotFoundError`

        :param str resource_name: the resource name.
        :param str parent_id: the resource parent.

        :param str object_id: unique identifier of the object
        :param bool with_deleted: track deleted object with a tombstone

        :returns: the deleted object, with minimal set of attributes.
        :rtype: dict
        """
        raise NotImplementedError

    def delete_all(
        self,
        resource_name,
        parent_id,
        filters=None,
        sorting=None,
        pagination_rules=None,
        limit=None,
        id_field=DEFAULT_ID_FIELD,
        with_deleted=True,
        modified_field=DEFAULT_MODIFIED_FIELD,
        deleted_field=DEFAULT_DELETED_FIELD,
    ):
        """Delete all objects in this `resource_name` for this `parent_id`.

        :param str resource_name: the resource name.
        :param str parent_id: the resource parent.

        :param filters: Optionnally filter the objects to delete.
        :type filters: list of :class:`kinto.core.storage.Filter`
        :param sorting: Optionnally sort the objects by attribute.
            Each sort instruction in this list refers to a field and a
            direction (negative means descending). All sort instructions are
            cumulative.
        :type sorting: list of :class:`kinto.core.storage.Sort`

        :param pagination_rules: Optionnally paginate the deletion of objects.
            This list of rules aims to reduce the set of objects to the current
            page. A rule is a list of filters (see `filters` parameter),
            and all rules are combined using *OR*.
        :type pagination_rules: list of list of
            :class:`kinto.core.storage.Filter`

        :param int limit: Optionnally limit the number of objects to be
            deleted.

        :param bool with_deleted: track deleted objects with a tombstone

        :returns: the list of deleted objects, with minimal set of attributes.
        :rtype: list
        """
        raise NotImplementedError

    def purge_deleted(
        self,
        resource_name,
        parent_id,
        before=None,
        id_field=DEFAULT_ID_FIELD,
        modified_field=DEFAULT_MODIFIED_FIELD,
    ):
        """Delete all deleted object tombstones in this `resource_name`
        for this `parent_id`.

        :param str resource_name: the resource name.
        :param str parent_id: the resource parent.

        :param int before: Optionnal timestamp to limit deletion (exclusive)

        :returns: The number of deleted objects.
        :rtype: int

        """
        raise NotImplementedError

    @deprecate_kwargs({"collection_id": "resource_name"})
    def get_all(self, *args, **kwargs):
        """Legacy method to support code that relied on the old API where the storage's
        get_all() would return a tuple of (<list of objects paginated>, <count of all>).
        Since then, we're being more explicit and expecting the client to deliberately
        decide if they need a paginated list or a count.

        This method exists solely to make the transition easier.
        """
        warnings.warn("Use either self.list_all() or self.count_all()", DeprecationWarning)
        list_ = self.list_all(*args, **kwargs)
        kwargs.pop("pagination_rules", None)
        kwargs.pop("limit", None)
        kwargs.pop("sorting", None)
        kwargs.pop("include_deleted", None)
        count = self.count_all(*args, **kwargs)
        return (list_, count)

    def list_all(
        self,
        resource_name,
        parent_id,
        filters=None,
        sorting=None,
        pagination_rules=None,
        limit=None,
        include_deleted=False,
        id_field=DEFAULT_ID_FIELD,
        modified_field=DEFAULT_MODIFIED_FIELD,
        deleted_field=DEFAULT_DELETED_FIELD,
    ):
        """Retrieve all objects in this `resource_name` for this `parent_id`.

        :param str resource_name: the resource name.

        :param str parent_id: the resource parent, possibly
            containing a wildcard '*'. (This can happen when
            implementing "administrator" operations on a Resource,
            for example, like ``kinto.plugins.accounts``.)

        :param filters: Optionally filter the objects by their attribute.
            Each filter in this list is a tuple of a field, a value and a
            comparison (see `kinto.core.utils.COMPARISON`). All filters
            are combined using *AND*.
        :type filters: list of :class:`kinto.core.storage.Filter`

        :param sorting: Optionnally sort the objects by attribute.
            Each sort instruction in this list refers to a field and a
            direction (negative means descending). All sort instructions are
            cumulative.
        :type sorting: list of :class:`kinto.core.storage.Sort`

        :param pagination_rules: Optionnally paginate the list of objects.
            This list of rules aims to reduce the set of objects to the current
            page. A rule is a list of filters (see `filters` parameter),
            and all rules are combined using *OR*.
        :type pagination_rules: list of list of
            :class:`kinto.core.storage.Filter`

        :param int limit: Optionnally limit the number of objects to be
            retrieved.

        :param bool include_deleted: Optionnally include the deleted objects
            that match the filters.

        :returns: the limited list of objects of
            matching objects in the resource (deleted ones excluded).
        :rtype: list
        """
        raise NotImplementedError

    def count_all(
        self,
        resource_name,
        parent_id,
        filters=None,
        id_field=DEFAULT_ID_FIELD,
        modified_field=DEFAULT_MODIFIED_FIELD,
        deleted_field=DEFAULT_DELETED_FIELD,
    ):
        """Return a count of all objects in this `resource_name` for this `parent_id`.

        :param str resource_name: the resource name.
        :param str parent_id: the parent resource, possibly
            containing a wildcard '*'. (This can happen when
            implementing "administrator" operations on a UserResource,
            for example.)
        :param filters: Optionally filter the objects by their attribute.
            Each filter in this list is a tuple of a field, a value and a
            comparison (see `kinto.core.utils.COMPARISON`). All filters
            are combined using *AND*.
        :type filters: list of :class:`kinto.core.storage.Filter`
        :returns: the total number of matching objects in the resource (deleted ones excluded).
        :rtype: int
        """
        raise NotImplementedError

    def collection_timestamp(self, collection_id, parent_id):
        message = "`collection_timestamp()` is deprecated, use `resource_timestamp()` instead."
        warnings.warn(message, DeprecationWarning)
        return self.resource_timestamp(resource_name=collection_id, parent_id=parent_id)


def heartbeat(backend):
    def ping(request):
        """Test that storage is operational.

        :param request: current request object
        :type request: :class:`~pyramid:pyramid.request.Request`
        :returns: ``True`` is everything is ok, ``False`` otherwise.
        :rtype: bool
        """
        try:
            storage_kw = dict(resource_name=_HEARTBEAT_RESOURCE_NAME, parent_id=_HEART_PARENT_ID)
            if asbool(request.registry.settings.get("readonly")):
                # Do not try to write in readonly mode.
                backend.get_all(**storage_kw)
            else:
                if random.SystemRandom().random() < _HEARTBEAT_DELETE_RATE:
                    backend.delete_all(**storage_kw)
                    backend.purge_deleted(**storage_kw)  # Kinto/kinto#985
                else:
                    backend.create(obj=_HEARTBEAT_OBJECT, **storage_kw)
            return True
        except Exception:
            logger.exception("Heartbeat Error")
            return False

    return ping



from unittest import mock

from pyramid import httpexceptions

from kinto.core.testing import unittest

from ..support import BaseWebTest

MINIMALIST_OBJECT = {"name": "Champignon"}


class CORSOriginHeadersTest(BaseWebTest, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.headers["Origin"] = "notmyidea.org"

    def setUp(self):
        super().setUp()
        body = {"data": MINIMALIST_OBJECT}
        response = self.app.post_json(self.plural_url, body, headers=self.headers, status=201)
        self.obj = response.json["data"]

    def test_can_be_configured_from_settings(self):
        app = self.make_app({"cors_origins": "*.daybed.io"})
        headers = {**self.headers, "Origin": "demo.daybed.io"}
        resp = app.get(self.plural_url, headers=headers)
        self.assertEqual(resp.headers["Access-Control-Allow-Origin"], "demo.daybed.io")

    def test_present_on_hello(self):
        response = self.app.get("/", headers=self.headers, status=200)
        self.assertIn("Access-Control-Allow-Origin", response.headers)

    def test_present_on_single_object(self):
        response = self.app.get(self.get_item_url(), headers=self.headers)
        self.assertIn("Access-Control-Allow-Origin", response.headers)

    def test_present_on_deletion(self):
        response = self.app.delete(self.get_item_url(), headers=self.headers)
        self.assertIn("Access-Control-Allow-Origin", response.headers)

    def test_present_on_unknown_url(self):
        response = self.app.get("/unknown", headers=self.headers, status=404)
        self.assertEqual(response.headers["Access-Control-Allow-Origin"], "notmyidea.org")

    def test_not_present_on_unknown_url_if_setting_does_not_match(self):
        with mock.patch.dict(self.app.app.registry.settings, [("cors_origins", "daybed.io")]):
            response = self.app.get("/unknown", headers=self.headers, status=404)
            self.assertNotIn("Access-Control-Allow-Origin", response.headers)

    def test_present_on_unknown_object(self):
        url = self.get_item_url("1cea99eb-5e3d-44ad-a53a-2fb68473b538")
        response = self.app.get(url, headers=self.headers, status=404)
        self.assertIn("Access-Control-Allow-Origin", response.headers)

    def test_present_on_invalid_object_update(self):
        body = {"data": {"name": 42}}
        response = self.app.patch_json(self.get_item_url(), body, headers=self.headers, status=400)
        self.assertIn("Access-Control-Allow-Origin", response.headers)

    def test_present_on_successful_creation(self):
        body = {"data": MINIMALIST_OBJECT}
        response = self.app.post_json(self.plural_url, body, headers=self.headers, status=201)
        self.assertIn("Access-Control-Allow-Origin", response.headers)

    def test_present_on_invalid_object_creation(self):
        body = {"name": 42}
        response = self.app.post_json(self.plural_url, body, headers=self.headers, status=400)
        self.assertIn("Access-Control-Allow-Origin", response.headers)

    def test_present_on_readonly_update(self):
        with mock.patch("tests.core.testapp.views.MushroomSchema.is_readonly", return_value=True):
            body = {"data": {"name": "Amanite"}}
            response = self.app.patch_json(
                self.get_item_url(), body, headers=self.headers, status=400
            )
        self.assertEqual(response.json["message"], "Cannot modify name")
        self.assertIn("Access-Control-Allow-Origin", response.headers)

    def test_present_on_unauthorized(self):
        headers = {**self.headers}
        headers.pop("Authorization", None)
        body = {"data": MINIMALIST_OBJECT}
        response = self.app.post_json(self.plural_url, body, headers=headers, status=401)
        self.assertIn("Access-Control-Allow-Origin", response.headers)

    def test_present_on_internal_error(self):
        with mock.patch("kinto.core.resource.Resource._extract_filters", side_effect=ValueError):
            response = self.app.get("/mushrooms", headers=self.headers, status=500)
        self.assertIn("Access-Control-Allow-Origin", response.headers)

    def test_present_on_http_error(self):
        with mock.patch(
            "kinto.core.resource.Resource._extract_filters",
            side_effect=httpexceptions.HTTPPaymentRequired,
        ):
            response = self.app.get("/mushrooms", headers=self.headers, status=402)
        self.assertIn("Access-Control-Allow-Origin", response.headers)


class CORSExposeHeadersTest(BaseWebTest, unittest.TestCase):
    def assert_expose_headers(self, method, path, allowed_headers, body=None, status=None):
        headers = {**self.headers, "Origin": "lolnet.org"}
        http_method = getattr(self.app, method.lower())
        kwargs = dict(headers=headers)
        if status:
            kwargs["status"] = status
        if method == "HEAD":
            # TestApp.head can't take a 'params' and its first and only positional argument.
            # So we need an ad-hoc signature for TestApp.head.
            response = http_method(path, **kwargs)
        else:
            response = http_method(path, body, **kwargs)
        exposed_headers = response.headers["Access-Control-Expose-Headers"]
        exposed_headers = [x.strip() for x in exposed_headers.split(",")]
        self.assertEqual(sorted(allowed_headers), sorted(exposed_headers))
        return response

    def test_plural_get_exposes_every_possible_header(self):
        self.assert_expose_headers(
            "GET",
            self.plural_url,
            [
                "Alert",
                "Backoff",
                "ETag",
                "Last-Modified",
                "Next-Page",
                "Retry-After",
                "Content-Length",
                "Content-Type",
                "Cache-Control",
                "Expires",
                "Pragma",
            ],
        )

    def test_plural_head_exposes_every_possible_header(self):
        self.assert_expose_headers(
            "HEAD",
            self.plural_url,
            [
                "Alert",
                "Backoff",
                "ETag",
                "Last-Modified",
                "Retry-After",
                "Total-Objects",
                "Total-Records",
                "Content-Length",
                "Content-Type",
                "Cache-Control",
                "Expires",
                "Next-Page",
                "Pragma",
            ],
        )

    def test_hello_endpoint_exposes_only_minimal_set_of_headers(self):
        self.assert_expose_headers(
            "GET", "/", ["Alert", "Backoff", "Retry-After", "Content-Type", "Content-Length"]
        )

    def test_object_get_exposes_only_used_headers(self):
        body = {"data": MINIMALIST_OBJECT}
        resp = self.app.post_json(self.plural_url, body, headers=self.headers, status=201)
        object_url = self.get_item_url(resp.json["data"]["id"])
        self.assert_expose_headers(
            "GET",
            object_url,
            [
                "Alert",
                "Backoff",
                "ETag",
                "Retry-After",
                "Last-Modified",
                "Content-Length",
                "Content-Type",
                "Cache-Control",
                "Expires",
                "Pragma",
            ],
        )

    def test_object_post_exposes_only_minimal_set_of_headers(self):
        body = {"data": MINIMALIST_OBJECT}
        self.assert_expose_headers(
            "POST_JSON",
            "/mushrooms",
            ["Alert", "Backoff", "Retry-After", "Content-Length", "Content-Type"],
            body=body,
        )

    def test_present_on_bad_id_400_errors(self):
        body = {"data": {"name": "Amanite"}}
        self.assert_expose_headers(
            "PUT_JSON",
            "/mushrooms/wrong=ids",
            ["Alert", "Backoff", "Retry-After", "Content-Length", "Content-Type"],
            body=body,
            status=400,
        )

    def test_present_on_unknown_url(self):
        self.assert_expose_headers(
            "PUT_JSON",
            "/unknown",
            ["Alert", "Backoff", "Retry-After", "Content-Length", "Content-Type"],
            status=404,
        )


class CORSMaxAgeTest(BaseWebTest, unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        super().setUpClass()
        cls.headers.update({"Origin": "lolnet.org", "Access-Control-Request-Method": "GET"})

    def test_cors_max_age_is_3600_seconds_by_default(self):
        app = self.make_app()
        resp = app.options("/", headers=self.headers)
        self.assertEqual(int(resp.headers["Access-Control-Max-Age"]), 3600)

    def test_cors_max_age_can_be_specified_in_settings(self):
        app = self.make_app({"cors_max_age_seconds": "42"})
        resp = app.options("/", headers=self.headers)
        self.assertEqual(int(resp.headers["Access-Control-Max-Age"]), 42)

    def test_cors_max_age_is_disabled_if_unset(self):
        app = self.make_app({"cors_max_age_seconds": ""})
        resp = app.options("/", headers=self.headers)
        self.assertNotIn("Access-Control-Max-Age", resp.headers)



import logging
import random
import re
import warnings
from datetime import datetime

from dateutil import parser as dateparser
from pyramid.events import NewRequest, NewResponse
from pyramid.exceptions import ConfigurationError
from pyramid.httpexceptions import HTTPBadRequest, HTTPGone, HTTPTemporaryRedirect
from pyramid.interfaces import IAuthenticationPolicy
from pyramid.renderers import JSON as JSONRenderer
from pyramid.response import Response
from pyramid.security import NO_PERMISSION_REQUIRED
from pyramid.settings import asbool, aslist
from pyramid_multiauth import MultiAuthenticationPolicy, MultiAuthPolicySelected

from kinto.core import cache, errors, permission, storage, utils
from kinto.core.events import ACTIONS, ResourceChanged, ResourceRead

try:
    import newrelic.agent
except ImportError:  # pragma: no cover
    newrelic = None
try:
    from werkzeug.middleware.profiler import ProfilerMiddleware
except ImportError:  # pragma: no cover
    pass


logger = logging.getLogger(__name__)
summary_logger = logging.getLogger("request.summary")


def setup_request_bound_data(config):
    """Attach custom data on request object, and share it with parent
    requests during batch."""

    def attach_bound_data(request):
        parent = getattr(request, "parent", None)
        return parent.bound_data if parent else {}

    config.add_request_method(attach_bound_data, name="bound_data", reify=True)


def setup_json_serializer(config):
    import requests
    import webob

    # Monkey patch to use ujson
    webob.request.json = utils.json
    requests.models.json = utils.json

    # Override json renderer using ujson
    renderer = JSONRenderer(serializer=utils.json_serializer)
    config.add_renderer("json", renderer)


def setup_csp_headers(config):
    """Content-Security-Policy HTTP response header helps reduce XSS risks on
    modern browsers by declaring, which dynamic resources are allowed to load.
    On APIs, we disable everything.
    """
    disable_all = "default-src 'none'; frame-ancestors 'none'; base-uri 'none'; "

    def on_new_response(event):
        event.response.headers.setdefault("Content-Security-Policy", disable_all)

    config.add_subscriber(on_new_response, NewResponse)


def setup_version_redirection(config):
    """Add a view which redirects to the current version of the API.
    """
    settings = config.get_settings()
    redirect_enabled = settings["version_prefix_redirect_enabled"]
    version_prefix_redirection_enabled = asbool(redirect_enabled)

    route_prefix = config.route_prefix
    config.registry.route_prefix = route_prefix

    # Redirect to the current version of the API if the prefix isn't used.
    # Do not redirect if kinto.version_prefix_redirect_enabled is set to
    # False.
    if not version_prefix_redirection_enabled:
        return

    def _redirect_to_version_view(request):
        if request.method.lower() == "options":
            # CORS responses should always have status 200.
            return utils.reapply_cors(request, Response())

        querystring = request.url[(request.url.rindex(request.path) + len(request.path)) :]
        redirect = f"/{route_prefix}{request.path}{querystring}"
        raise HTTPTemporaryRedirect(redirect)

    # Disable the route prefix passed by the app.
    config.route_prefix = None

    config.add_route(name="redirect_to_version", pattern=r"/{path:(?!v[0-9]+)[^\r\n]*}")

    config.add_view(
        view=_redirect_to_version_view,
        route_name="redirect_to_version",
        permission=NO_PERMISSION_REQUIRED,
    )

    config.route_prefix = route_prefix


def setup_authentication(config):
    """Let pyramid_multiauth manage authentication and authorization
    from configuration.
    """
    config.include("pyramid_multiauth")
    settings = config.get_settings()

    policies = aslist(settings["multiauth.policies"])
    if "basicauth" in policies:
        config.include("kinto.core.authentication")

    # Track policy used, for prefixing user_id and for logging.
    def on_policy_selected(event):
        request = event.request
        # If the authn policy has a name attribute, we use it rather
        # than using the one specified in settings.
        authn_type = getattr(event.policy, "name", event.policy_name.lower())
        request.authn_type = authn_type
        request.selected_userid = event.userid
        # Add authentication info to context.
        request.log_context(uid=event.userid, authn_type=authn_type)

    config.add_subscriber(on_policy_selected, MultiAuthPolicySelected)


def setup_backoff(config):
    """Attach HTTP requests/responses objects.

    This is useful to attach objects to the request object for easier
    access, and to pre-process responses.
    """

    def on_new_response(event):
        # Add backoff in response headers.
        backoff = config.registry.settings["backoff"]
        if backoff is not None:
            backoff_percentage = config.registry.settings["backoff_percentage"]
            if backoff_percentage is not None:
                if random.random() < (float(backoff_percentage) / 100.0):
                    event.response.headers["Backoff"] = str(backoff)
            else:
                event.response.headers["Backoff"] = str(backoff)

    config.add_subscriber(on_new_response, NewResponse)


def setup_requests_scheme(config):
    """Force server scheme, host and port at the application level."""
    settings = config.get_settings()

    http_scheme = settings["http_scheme"]
    http_host = settings["http_host"]

    def on_new_request(event):
        if http_scheme:
            event.request.scheme = http_scheme
        if http_host:
            event.request.host = http_host

    if http_scheme or http_host:
        config.add_subscriber(on_new_request, NewRequest)


def setup_deprecation(config):
    config.add_tween("kinto.core.initialization._end_of_life_tween_factory")


def _end_of_life_tween_factory(handler, registry):
    """Pyramid tween to handle service end of life."""
    deprecation_msg = "The service you are trying to connect no longer exists" " at this location."

    def eos_tween(request):
        eos_date = registry.settings["eos"]
        eos_url = registry.settings["eos_url"]
        eos_message = registry.settings["eos_message"]
        if not eos_date:
            return handler(request)

        eos_date = dateparser.parse(eos_date)
        if eos_date > datetime.now():
            code = "soft-eol"
            request.response = handler(request)
        else:
            code = "hard-eol"
            request.response = errors.http_error(
                HTTPGone(), errno=errors.ERRORS.SERVICE_DEPRECATED, message=deprecation_msg
            )

        errors.send_alert(request, eos_message, url=eos_url, code=code)
        return utils.reapply_cors(request, request.response)

    return eos_tween


def setup_storage(config):
    settings = config.get_settings()

    # Id generators by resource name.
    config.registry.id_generators = {}
    for key, value in settings.items():
        m = re.match(r"^([^_]*)_?id_generator", key)
        if m is None:
            continue
        resource_name = m.group(1)
        id_generator = config.maybe_dotted(value)
        config.registry.id_generators[resource_name] = id_generator()

    storage_mod = settings["storage_backend"]
    if not storage_mod:
        return

    storage_mod = config.maybe_dotted(storage_mod)
    backend = storage_mod.load_from_config(config)
    if not isinstance(backend, storage.StorageBase):
        raise ConfigurationError(f"Invalid storage backend: {backend}")
    config.registry.storage = backend

    heartbeat = storage.heartbeat(backend)
    config.registry.heartbeats["storage"] = heartbeat


def setup_permission(config):
    settings = config.get_settings()
    permission_mod = settings["permission_backend"]
    if not permission_mod:
        return

    permission_mod = config.maybe_dotted(permission_mod)
    backend = permission_mod.load_from_config(config)
    if not isinstance(backend, permission.PermissionBase):
        raise ConfigurationError(f"Invalid permission backend: {backend}")
    config.registry.permission = backend

    heartbeat = permission.heartbeat(backend)
    config.registry.heartbeats["permission"] = heartbeat


def setup_cache(config):
    settings = config.get_settings()
    cache_mod = settings["cache_backend"]
    if not cache_mod:
        return

    cache_mod = config.maybe_dotted(cache_mod)
    backend = cache_mod.load_from_config(config)
    if not isinstance(backend, cache.CacheBase):
        raise ConfigurationError(f"Invalid cache backend: {backend}")
    config.registry.cache = backend

    heartbeat = cache.heartbeat(backend)
    config.registry.heartbeats["cache"] = heartbeat


def setup_statsd(config):
    settings = config.get_settings()
    config.registry.statsd = None

    if settings["statsd_url"]:
        statsd_mod = settings["statsd_backend"]
        statsd_mod = config.maybe_dotted(statsd_mod)
        client = statsd_mod.load_from_config(config)

        config.registry.statsd = client

        client.watch_execution_time(config.registry.cache, prefix="backend")
        client.watch_execution_time(config.registry.storage, prefix="backend")
        client.watch_execution_time(config.registry.permission, prefix="backend")

        # Commit so that configured policy can be queried.
        config.commit()
        policy = config.registry.queryUtility(IAuthenticationPolicy)
        if isinstance(policy, MultiAuthenticationPolicy):
            for name, subpolicy in policy.get_policies():
                client.watch_execution_time(subpolicy, prefix="authentication", classname=name)
        else:
            client.watch_execution_time(policy, prefix="authentication")

        def on_new_response(event):
            request = event.request

            # Count unique users.
            user_id = request.prefixed_userid
            if user_id:
                # Get rid of colons in metric packet (see #1282).
                user_id = user_id.replace(":", ".")
                client.count("users", unique=user_id)

            # Count authentication verifications.
            if hasattr(request, "authn_type"):
                client.count(f"authn_type.{request.authn_type}")

            # Count view calls.
            service = request.current_service
            if service:
                client.count(f"view.{service.name}.{request.method}")

        config.add_subscriber(on_new_response, NewResponse)

        return client


def install_middlewares(app, settings):
    "Install a set of middlewares defined in the ini file on the given app."
    # Setup new-relic.
    if settings.get("newrelic_config"):
        ini_file = settings["newrelic_config"]
        env = settings["newrelic_env"]
        newrelic.agent.initialize(ini_file, env)
        app = newrelic.agent.WSGIApplicationWrapper(app)

    # Adds the Werkzeug profiler.
    if asbool(settings.get("profiler_enabled")):
        profile_dir = settings["profiler_dir"]
        app = ProfilerMiddleware(app, profile_dir=profile_dir, restrictions=("*kinto.core*"))

    return app


def setup_logging(config):
    """Setup structured logging, and emit `request.summary` event on each
    request, as recommended by Mozilla Services standard:

    * https://mana.mozilla.org/wiki/display/CLOUDSERVICES/Logging+Standard
    * http://12factor.net/logs
    """

    def on_new_request(event):
        request = event.request
        # Save the time the request was received by the server.
        event.request._received_at = utils.msec_time()

        try:
            # Pyramid fails if the URL contains invalid UTF-8 characters.
            request_path = event.request.path
        except UnicodeDecodeError:
            raise errors.http_error(
                HTTPBadRequest(),
                errno=errors.ERRORS.INVALID_PARAMETERS,
                message="Invalid URL path.",
            )

        request.log_context(
            agent=request.headers.get("User-Agent"),
            path=request_path,
            method=request.method,
            lang=request.headers.get("Accept-Language"),
            errno=0,
        )
        qs = dict(errors.request_GET(request))
        if qs:
            request.log_context(querystring=qs)

        if summary_logger.level == logging.DEBUG:
            request.log_context(headers=dict(request.headers), body=request.body)

    config.add_subscriber(on_new_request, NewRequest)

    def on_new_response(event):
        response = event.response
        request = event.request

        # Compute the request processing time in msec (-1 if unknown)
        current = utils.msec_time()
        duration = current - getattr(request, "_received_at", current - 1)
        isotimestamp = datetime.fromtimestamp(current / 1000).isoformat()

        # Bind infos for request summary logger.
        request.log_context(time=isotimestamp, code=response.status_code, t=duration)

        if summary_logger.level == logging.DEBUG:
            request.log_context(response=dict(headers=dict(response.headers), body=response.body))

        try:
            # If error response, bind errno.
            request.log_context(errno=response.errno)
        except AttributeError:
            pass

        if not hasattr(request, "parent"):
            # Ouput application request summary.
            summary_logger.info("", extra=request.log_context())

    config.add_subscriber(on_new_response, NewResponse)


class EventActionFilter:
    def __init__(self, actions, config):
        actions = ACTIONS.from_string_list(actions)
        self.actions = [action.value for action in actions]

    def phash(self):
        return f"for_actions = {','.join(self.actions)}"

    def __call__(self, event):
        action = event.payload.get("action")
        return not action or action in self.actions


class EventResourceFilter:
    def __init__(self, resources, config):
        self.resources = resources

    def phash(self):
        return f"for_resources = {','.join(self.resources)}"

    def __call__(self, event):
        resource = event.payload.get("resource_name")
        return not resource or not self.resources or resource in self.resources


def setup_listeners(config):
    # Register basic subscriber predicates, to filter events.
    config.add_subscriber_predicate("for_actions", EventActionFilter)
    config.add_subscriber_predicate("for_resources", EventResourceFilter)

    write_actions = (ACTIONS.CREATE, ACTIONS.UPDATE, ACTIONS.DELETE)
    settings = config.get_settings()
    settings_prefix = settings.get("settings_prefix", "")
    listeners = aslist(settings["event_listeners"])

    for name in listeners:
        logger.info(f"Setting up '{name}' listener")
        prefix = f"event_listeners.{name}."

        try:
            listener_mod = config.maybe_dotted(name)
            prefix = f"event_listeners.{name.split('.')[-1]}."
            listener = listener_mod.load_from_config(config, prefix)
        except (ImportError, AttributeError):
            module_setting = prefix + "use"
            # Read from ENV or settings.
            module_value = utils.read_env(
                f"{settings_prefix}.{module_setting}", settings.get(module_setting)
            )
            listener_mod = config.maybe_dotted(module_value)
            listener = listener_mod.load_from_config(config, prefix)

        # If StatsD is enabled, monitor execution time of listeners.
        if getattr(config.registry, "statsd", None):
            statsd_client = config.registry.statsd
            key = f"listeners.{name}"
            listener = statsd_client.timer(key)(listener.__call__)

        # Optional filter by event action.
        actions_setting = prefix + "actions"
        # Read from ENV or settings.
        actions_value = utils.read_env(
            f"{settings_prefix}.{actions_setting}", settings.get(actions_setting, "")
        )
        actions = aslist(actions_value)
        if len(actions) > 0:
            actions = ACTIONS.from_string_list(actions)
        else:
            actions = write_actions

        # Optional filter by event resource name.
        resource_setting = prefix + "resources"
        # Read from ENV or settings.
        resource_value = utils.read_env(
            f"{settings_prefix}.{resource_setting}", settings.get(resource_setting, "")
        )
        resource_names = aslist(resource_value)

        # Pyramid event predicates.
        options = dict(for_actions=actions, for_resources=resource_names)

        if ACTIONS.READ in actions:
            config.add_subscriber(listener, ResourceRead, **options)
            actions = [a for a in actions if a != ACTIONS.READ]

        if len(actions) > 0:
            config.add_subscriber(listener, ResourceChanged, **options)


def load_default_settings(config, default_settings):
    """Read settings provided in Paste ini file, set default values and
    replace if defined as environment variable.
    """
    settings = config.get_settings()

    settings_prefix = settings["settings_prefix"]

    def _prefixed_keys(key):
        unprefixed = key
        if key.startswith(settings_prefix + "."):
            unprefixed = key.split(".", 1)[1]
        project_prefix = f"{settings_prefix}.{unprefixed}"
        return unprefixed, project_prefix

    # Fill settings with default values if not defined.
    for key, default_value in sorted(default_settings.items()):
        unprefixed, project_prefix = keys = _prefixed_keys(key)
        is_defined = len(set(settings.keys()).intersection(set(keys))) > 0
        if not is_defined:
            settings[unprefixed] = default_value

    for key, value in sorted(settings.items()):
        value = utils.native_value(value)
        unprefixed, project_prefix = keys = _prefixed_keys(key)

        # Fail if not only one is defined.
        defined = set(settings.keys()).intersection(set(keys))
        distinct_values = set([str(settings[d]) for d in defined])

        if len(defined) > 1 and len(distinct_values) > 1:
            names = "', '".join(defined)
            raise ValueError(f"Settings '{names}' are in conflict.")

        # Override settings from OS env values.
        # e.g. HTTP_PORT, READINGLIST_HTTP_PORT, KINTO_HTTP_PORT
        from_env = utils.read_env(unprefixed, value)
        from_env = utils.read_env(project_prefix, from_env)

        settings[unprefixed] = from_env

    config.add_settings(settings)


def initialize(config, version=None, settings_prefix="", default_settings=None):
    """Initialize kinto.core with the given configuration, version and project
    name.

    This will basically include kinto.core in Pyramid and set route prefix
    based on the specified version.

    :param config: Pyramid configuration
    :type config: :class:`~pyramid:pyramid.config.Configurator`
    :param str version: Current project version (e.g. '0.0.1') if not defined
        in application settings.
    :param str settings_prefix: Project name if not defined
        in application settings.
    :param dict default_settings: Override kinto.core default settings values.
    """
    from kinto.core import DEFAULT_SETTINGS

    settings = config.get_settings()

    settings_prefix = (
        settings.pop("kinto.settings_prefix", settings.get("settings_prefix")) or settings_prefix
    )
    settings["settings_prefix"] = settings_prefix
    if not settings_prefix:
        warnings.warn("No value specified for `settings_prefix`")

    kinto_core_defaults = {**DEFAULT_SETTINGS}

    if default_settings:
        kinto_core_defaults.update(default_settings)

    load_default_settings(config, kinto_core_defaults)

    http_scheme = settings["http_scheme"]
    if http_scheme != "https":
        warnings.warn("HTTPS is not enabled")

    # Override project version from settings.
    project_version = settings.get("project_version") or version
    if not project_version:
        error_msg = f"Invalid project version: {project_version}"
        raise ConfigurationError(error_msg)
    settings["project_version"] = project_version = str(project_version)

    # HTTP API version.
    http_api_version = settings.get("http_api_version")
    if http_api_version is None:
        # The API version is derivated from the module version if not provided.
        http_api_version = ".".join(project_version.split(".")[0:2])
    settings["http_api_version"] = http_api_version = str(http_api_version)
    api_version = f"v{http_api_version.split('.')[0]}"

    # Include kinto.core views with the correct api version prefix.
    config.include("kinto.core", route_prefix=api_version)
    config.route_prefix = api_version



import colander
from pyramid.events import subscriber

from kinto.core import resource, utils
from kinto.core.events import ACTIONS, ResourceChanged
from kinto.schema_validation import JSONSchemaMapping, validate_from_bucket_schema_or_400


class CollectionSchema(resource.ResourceSchema):
    schema = JSONSchemaMapping(missing=colander.drop)
    cache_expires = colander.SchemaNode(colander.Int(), missing=colander.drop)


@resource.register(
    name="collection",
    plural_path="/buckets/{{bucket_id}}/collections",
    object_path="/buckets/{{bucket_id}}/collections/{{id}}",
)
class Collection(resource.Resource):
    schema = CollectionSchema
    permissions = ("read", "write", "record:create")

    def get_parent_id(self, request):
        bucket_id = request.matchdict["bucket_id"]
        parent_id = utils.instance_uri(request, "bucket", id=bucket_id)
        return parent_id

    def process_object(self, new, old=None):
        """Additional collection schema validation from bucket, if any."""
        new = super().process_object(new, old)

        # Remove internal and auto-assigned fields.
        internal_fields = (self.model.modified_field, self.model.permissions_field)
        validate_from_bucket_schema_or_400(
            new,
            resource_name="collection",
            request=self.request,
            ignore_fields=internal_fields,
            id_field=self.model.id_field,
        )
        return new


@subscriber(ResourceChanged, for_resources=("collection",), for_actions=(ACTIONS.DELETE,))
def on_collections_deleted(event):
    """Some collections were deleted, delete records.
    """
    storage = event.request.registry.storage
    permission = event.request.registry.permission

    for change in event.impacted_objects:
        collection = change["old"]
        bucket_id = event.payload["bucket_id"]
        parent_id = utils.instance_uri(
            event.request, "collection", bucket_id=bucket_id, id=collection["id"]
        )
        storage.delete_all(resource_name=None, parent_id=parent_id, with_deleted=False)
        storage.purge_deleted(resource_name=None, parent_id=parent_id)
        permission.delete_object_permissions(parent_id + "/*")



from bravado_core.response import validate_response

from kinto.core import testing

from ..support import MINIMALIST_BUCKET, MINIMALIST_COLLECTION, MINIMALIST_GROUP, MINIMALIST_RECORD
from .support import OpenAPITest


class OpenAPIObjectErrorResponsesTest(OpenAPITest):
    def setUp(self):
        super().setUp()

        self.bucket = self.app.put_json(
            "/buckets/b1", MINIMALIST_BUCKET, headers=self.headers
        ).json

        self.group = self.app.put_json(
            "/buckets/b1/groups/g1", MINIMALIST_GROUP, headers=self.headers
        ).json

        self.collection = self.app.put_json(
            "/buckets/b1/collections/c1", MINIMALIST_COLLECTION, headers=self.headers
        ).json

        self.record = self.app.put_json(
            "/buckets/b1/collections/c1/records/r1", MINIMALIST_RECORD, headers=self.headers
        ).json

    def test_object_get_304(self):
        headers = {
            **self.headers,
            "If-None-Match": '"{}"'.format(self.bucket["data"]["last_modified"]),
        }
        response = self.app.get("/buckets/b1", headers=headers, status=304)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].get_bucket
        schema = self.spec.deref(op.op_spec["responses"]["304"])
        validate_response(schema, op, response)

    def test_object_get_400(self):
        headers = {**self.headers, "If-None-Match": "aaa"}
        response = self.app.get("/buckets/b1", headers=headers, status=400)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].get_bucket
        schema = self.spec.deref(op.op_spec["responses"]["400"])
        validate_response(schema, op, response)

    def test_object_get_401(self):
        response = self.app.get("/buckets/b1", status=401)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].get_bucket
        schema = self.spec.deref(op.op_spec["responses"]["401"])
        validate_response(schema, op, response)

    def test_object_get_403(self):
        headers = {**self.headers, **testing.get_user_headers("aaa")}
        response = self.app.get("/buckets/b1", headers=headers, status=403)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].get_bucket
        schema = self.spec.deref(op.op_spec["responses"]["403"])
        validate_response(schema, op, response)

    def test_object_get_404(self):
        response = self.app.get("/buckets/b1/collections/col", headers=self.headers, status=404)
        response = self.cast_bravado_response(response)
        op = self.resources["Collections"].get_collection
        schema = self.spec.deref(op.op_spec["responses"]["404"])
        validate_response(schema, op, response)

    def test_object_get_406(self):
        headers = {**self.headers, "Accept": "text/html"}
        response = self.app.get("/buckets/b1", headers=headers, status=406)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].get_bucket
        schema = self.spec.deref(op.op_spec["responses"]["406"])
        validate_response(schema, op, response)

    def test_object_get_412(self):
        headers = {
            **self.headers,
            "If-Match": '"{}"'.format(self.bucket["data"]["last_modified"] - 1),
        }
        response = self.app.get("/buckets/b1", headers=headers, status=412)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].get_bucket
        schema = self.spec.deref(op.op_spec["responses"]["412"])
        validate_response(schema, op, response)

    def test_object_put_400(self):
        response = self.app.put_json("/buckets/b1", [], headers=self.headers, status=400)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].update_bucket
        schema = self.spec.deref(op.op_spec["responses"]["400"])
        validate_response(schema, op, response)

    def test_object_put_401(self):
        response = self.app.put_json("/buckets/b1", MINIMALIST_BUCKET, status=401)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].update_bucket
        schema = self.spec.deref(op.op_spec["responses"]["401"])
        validate_response(schema, op, response)

    def test_object_put_403(self):
        headers = {**self.headers, **testing.get_user_headers("aaa")}
        response = self.app.put_json("/buckets/b1", MINIMALIST_BUCKET, headers=headers, status=403)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].update_bucket
        schema = self.spec.deref(op.op_spec["responses"]["403"])
        validate_response(schema, op, response)

    def test_object_put_406(self):
        headers = {**self.headers, "Accept": "text/html"}
        response = self.app.put_json("/buckets/b1", MINIMALIST_BUCKET, headers=headers, status=406)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].update_bucket
        schema = self.spec.deref(op.op_spec["responses"]["406"])
        validate_response(schema, op, response)

    def test_object_put_412(self):
        headers = {
            **self.headers,
            "If-Match": '"{}"'.format(self.bucket["data"]["last_modified"] - 1),
        }
        response = self.app.put_json("/buckets/b1", MINIMALIST_BUCKET, headers=headers, status=412)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].update_bucket
        schema = self.spec.deref(op.op_spec["responses"]["412"])
        validate_response(schema, op, response)

    def test_object_put_415(self):
        headers = {**self.headers, "Content-Type": "text/html"}
        response = self.app.put_json("/buckets/b1", MINIMALIST_BUCKET, headers=headers, status=415)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].update_bucket
        schema = self.spec.deref(op.op_spec["responses"]["415"])
        validate_response(schema, op, response)

    def test_object_patch_400(self):
        response = self.app.patch_json("/buckets/b1", [], headers=self.headers, status=400)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].patch_bucket
        schema = self.spec.deref(op.op_spec["responses"]["400"])
        validate_response(schema, op, response)

    def test_object_patch_401(self):
        response = self.app.patch_json("/buckets/b1", MINIMALIST_BUCKET, status=401)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].patch_bucket
        schema = self.spec.deref(op.op_spec["responses"]["401"])
        validate_response(schema, op, response)

    def test_object_patch_403(self):
        headers = {**self.headers, **testing.get_user_headers("aaa")}
        response = self.app.patch_json(
            "/buckets/b1", MINIMALIST_BUCKET, headers=headers, status=403
        )
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].patch_bucket
        schema = self.spec.deref(op.op_spec["responses"]["403"])
        validate_response(schema, op, response)

    def test_object_patch_404(self):
        response = self.app.patch_json(
            "/buckets/b1/collections/col", MINIMALIST_COLLECTION, headers=self.headers, status=404
        )
        response = self.cast_bravado_response(response)
        op = self.resources["Collections"].patch_collection
        schema = self.spec.deref(op.op_spec["responses"]["404"])
        validate_response(schema, op, response)

    def test_object_patch_406(self):
        headers = {**self.headers, "Accept": "text/html"}
        response = self.app.patch_json(
            "/buckets/b1", MINIMALIST_BUCKET, headers=headers, status=406
        )
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].patch_bucket
        schema = self.spec.deref(op.op_spec["responses"]["406"])
        validate_response(schema, op, response)

    def test_object_patch_412(self):
        headers = {
            **self.headers,
            "If-Match": '"{}"'.format(self.bucket["data"]["last_modified"] - 1),
        }
        response = self.app.patch_json(
            "/buckets/b1", MINIMALIST_BUCKET, headers=headers, status=412
        )
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].patch_bucket
        schema = self.spec.deref(op.op_spec["responses"]["412"])
        validate_response(schema, op, response)

    def test_object_patch_415(self):
        headers = {**self.headers, "Content-Type": "text/html"}
        response = self.app.patch_json(
            "/buckets/b1", MINIMALIST_BUCKET, headers=headers, status=415
        )
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].patch_bucket
        schema = self.spec.deref(op.op_spec["responses"]["415"])
        validate_response(schema, op, response)

    def test_object_delete_400(self):
        headers = {**self.headers, "If-Match": "aaa"}
        response = self.app.delete("/buckets/b1", headers=headers, status=400)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].delete_bucket
        schema = self.spec.deref(op.op_spec["responses"]["400"])
        validate_response(schema, op, response)

    def test_object_delete_401(self):
        response = self.app.delete("/buckets/b1", status=401)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].delete_bucket
        schema = self.spec.deref(op.op_spec["responses"]["401"])
        validate_response(schema, op, response)

    def test_object_delete_403(self):
        headers = {**self.headers, **testing.get_user_headers("aaa")}
        response = self.app.delete("/buckets/b1", headers=headers, status=403)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].delete_bucket
        schema = self.spec.deref(op.op_spec["responses"]["403"])
        validate_response(schema, op, response)

    def test_object_delete_404(self):
        response = self.app.delete("/buckets/b1/collections/col", headers=self.headers, status=404)
        response = self.cast_bravado_response(response)
        op = self.resources["Collections"].delete_collection
        schema = self.spec.deref(op.op_spec["responses"]["404"])
        validate_response(schema, op, response)

    def test_object_delete_406(self):
        headers = {**self.headers, "Accept": "text/html"}
        response = self.app.delete("/buckets/b1", headers=headers, status=406)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].delete_bucket
        schema = self.spec.deref(op.op_spec["responses"]["406"])
        validate_response(schema, op, response)

    def test_object_delete_412(self):
        headers = {
            **self.headers,
            "If-Match": '"{}"'.format(self.bucket["data"]["last_modified"] - 1),
        }
        response = self.app.delete("/buckets/b1", headers=headers, status=412)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].delete_bucket
        schema = self.spec.deref(op.op_spec["responses"]["412"])
        validate_response(schema, op, response)

    def test_list_get_304(self):
        headers = {
            **self.headers,
            "If-None-Match": '"{}"'.format(self.bucket["data"]["last_modified"]),
        }
        response = self.app.get("/buckets", headers=headers, status=304)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].get_buckets
        schema = self.spec.deref(op.op_spec["responses"]["304"])
        validate_response(schema, op, response)

    def test_list_get_400(self):
        response = self.app.get("/buckets?_since=aaa", headers=self.headers, status=400)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].get_buckets
        schema = self.spec.deref(op.op_spec["responses"]["400"])
        validate_response(schema, op, response)

    def test_list_get_401(self):
        response = self.app.get("/buckets", status=401)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].get_buckets
        schema = self.spec.deref(op.op_spec["responses"]["401"])
        validate_response(schema, op, response)

    def test_list_get_403(self):
        headers = {**self.headers, **testing.get_user_headers("aaa")}
        response = self.app.get("/buckets/b1/collections", headers=headers, status=403)
        response = self.cast_bravado_response(response)
        op = self.resources["Collections"].get_collections
        schema = self.spec.deref(op.op_spec["responses"]["403"])
        validate_response(schema, op, response)

    def test_list_get_406(self):
        headers = {**self.headers, "Accept": "text/html"}
        response = self.app.get("/buckets", headers=headers, status=406)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].get_buckets
        schema = self.spec.deref(op.op_spec["responses"]["406"])
        validate_response(schema, op, response)

    def test_list_get_412(self):
        headers = {
            **self.headers,
            "If-Match": '"{}"'.format(self.bucket["data"]["last_modified"] - 1),
        }
        response = self.app.get("/buckets/b1", headers=headers, status=412)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].get_buckets
        schema = self.spec.deref(op.op_spec["responses"]["412"])
        validate_response(schema, op, response)

    def test_list_delete_401(self):
        response = self.app.delete("/buckets", status=401)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].delete_buckets
        schema = self.spec.deref(op.op_spec["responses"]["401"])
        validate_response(schema, op, response)

    def test_list_delete_403(self):
        headers = {**self.headers, **testing.get_user_headers("aaa")}
        response = self.app.delete("/buckets/b1/collections", headers=headers, status=403)
        response = self.cast_bravado_response(response)
        op = self.resources["Collections"].delete_collections
        schema = self.spec.deref(op.op_spec["responses"]["403"])
        validate_response(schema, op, response)

    def test_list_delete_405(self):
        # XXX
        pass

    def test_list_delete_406(self):
        headers = {**self.headers, "Accept": "text/html"}
        response = self.app.delete("/buckets", headers=headers, status=406)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].delete_buckets
        schema = self.spec.deref(op.op_spec["responses"]["406"])
        validate_response(schema, op, response)

    def test_list_delete_412(self):
        headers = {
            **self.headers,
            "If-Match": '"{}"'.format(self.bucket["data"]["last_modified"] - 1),
        }
        response = self.app.delete("/buckets", headers=headers, status=412)
        response = self.cast_bravado_response(response)
        op = self.resources["Buckets"].delete_buckets
        schema = self.spec.deref(op.op_spec["responses"]["412"])
        validate_response(schema, op, response)



from unittest import mock

import pytest

from kinto.core import DEFAULT_SETTINGS
from kinto.core.storage import (
    MISSING,
    Filter,
    Sort,
    StorageBase,
    exceptions,
    generators,
    memory,
    postgresql,
)
from kinto.core.storage.testing import StorageTest
from kinto.core.storage.utils import paginated
from kinto.core.testing import skip_if_no_postgresql, unittest
from kinto.core.utils import COMPARISON, sqlalchemy


class GeneratorTest(unittest.TestCase):
    def test_generic_has_mandatory_override(self):
        self.assertRaises(NotImplementedError, generators.Generator)

    def test_id_generator_must_respect_storage_backends(self):
        class Dumb(generators.Generator):
            def __call__(self):
                return "*" * 80

        self.assertRaises(ValueError, Dumb)

    def test_default_generator_allow_underscores_dash_alphabet(self):
        class Dumb(generators.Generator):
            def __call__(self):
                return "1234"

        generator = Dumb()
        self.assertTrue(generator.match("1_2_3-abc"))
        self.assertTrue(generator.match("abc_123"))
        self.assertFalse(generator.match("-1_2_3-abc"))
        self.assertFalse(generator.match("_1_2_3-abc"))

    def test_uuid_generator_pattern_allows_uuid_only(self):
        invalid_uuid = "XXX-00000000-0000-5000-a000-000000000000"
        generator = generators.UUID4()
        self.assertFalse(generator.match(invalid_uuid))

    def test_uuid_generator_pattern_is_not_restricted_to_uuid4(self):
        generator = generators.UUID4()
        valid_uuid = "fd800e8d-e8e9-3cac-f502-816cbed9bb6c"
        self.assertTrue(generator.match(valid_uuid))
        invalid_uuid4 = "00000000-0000-5000-a000-000000000000"
        self.assertTrue(generator.match(invalid_uuid4))
        invalid_uuid4 = "00000000-0000-4000-e000-000000000000"
        self.assertTrue(generator.match(invalid_uuid4))


class StorageBaseTest(unittest.TestCase):
    def setUp(self):
        self.storage = StorageBase()

    def test_mandatory_overrides(self):
        calls = [
            (self.storage.initialize_schema,),
            (self.storage.flush,),
            (self.storage.resource_timestamp, "", ""),
            (self.storage.create, "", "", {}),
            (self.storage.get, "", "", ""),
            (self.storage.update, "", "", "", {}),
            (self.storage.delete, "", "", ""),
            (self.storage.delete_all, "", ""),
            (self.storage.purge_deleted, "", ""),
            (self.storage.list_all, "", ""),
            (self.storage.count_all, "", ""),
        ]
        for call in calls:
            self.assertRaises(NotImplementedError, *call)

    def test_backend_error_message_provides_given_message_if_defined(self):
        error = exceptions.BackendError(message="Connection Error")
        self.assertEqual(str(error), "Connection Error")

    def test_backenderror_message_default_to_original_exception_message(self):
        error = exceptions.BackendError(ValueError("Pool Error"))
        self.assertEqual(str(error), "ValueError: Pool Error")


class MemoryBasedStorageTest(unittest.TestCase):
    def test_backend_raise_not_implemented_error(self):
        storage = memory.MemoryBasedStorage()
        with pytest.raises(NotImplementedError):
            storage.bump_and_store_timestamp("object", "/school/foo/students/bar")


class MemoryStorageTest(StorageTest, unittest.TestCase):
    backend = memory
    settings = {"storage_strict_json": True}

    def setUp(self):
        super().setUp()
        self.client_error_patcher = mock.patch.object(
            self.storage,
            "bump_and_store_timestamp",
            side_effect=exceptions.BackendError("Segmentation fault."),
        )

    def test_backend_error_provides_original_exception(self):
        pass

    def test_raises_backend_error_if_error_occurs_on_client(self):
        pass

    def test_backend_error_is_raised_anywhere(self):
        pass

    def test_backenderror_message_default_to_original_exception_message(self):
        pass

    def test_ping_logs_error_if_unavailable(self):
        pass


class LenientMemoryStorageTest(MemoryStorageTest):
    settings = {"storage_strict_json": False}

    def test_create_bytes_raises(self):
        data = {"steak": "haché".encode(encoding="utf-8")}
        self.assertIsInstance(data["steak"], bytes)
        self.assertIsNotNone(self.create_object(data))

    def test_update_bytes_raises(self):
        obj = self.create_object()

        new_object = {"steak": "haché".encode(encoding="utf-8")}
        self.assertIsInstance(new_object["steak"], bytes)

        self.assertIsNotNone(
            self.storage.update(object_id=obj["id"], obj=new_object, **self.storage_kw)
        )


@skip_if_no_postgresql
class PostgreSQLStorageTest(StorageTest, unittest.TestCase):
    backend = postgresql
    settings = {
        "storage_max_fetch_size": 10000,
        "storage_backend": "kinto.core.storage.postgresql",
        "storage_poolclass": "sqlalchemy.pool.StaticPool",
        "storage_url": "postgresql://postgres:postgres@localhost:5432/testdb",
        "storage_strict_json": True,
    }

    def setUp(self):
        super().setUp()
        self.client_error_patcher = mock.patch.object(
            self.storage.client, "session_factory", side_effect=sqlalchemy.exc.SQLAlchemyError
        )

    def test_number_of_fetched_objects_can_be_limited_in_settings(self):
        for i in range(4):
            self.create_object({"phone": "tel-{}".format(i)})

        results = self.storage.list_all(**self.storage_kw)
        self.assertEqual(len(results), 4)

        settings = {**self.settings, "storage_max_fetch_size": 2}
        config = self._get_config(settings=settings)
        limited = self.backend.load_from_config(config)

        results = limited.list_all(**self.storage_kw)
        self.assertEqual(len(results), 2)
        count = limited.count_all(**self.storage_kw)
        self.assertEqual(count, 4)

    def test_number_of_fetched_objects_is_per_page(self):
        for i in range(10):
            self.create_object({"number": i})

        settings = {**self.settings, "storage_max_fetch_size": 2}
        config = self._get_config(settings=settings)
        backend = self.backend.load_from_config(config)

        results = backend.list_all(
            pagination_rules=[[Filter("number", 1, COMPARISON.GT)]], **self.storage_kw
        )
        self.assertEqual(len(results), 2)
        count = backend.count_all(**self.storage_kw)
        self.assertEqual(count, 10)

    def test_connection_is_rolledback_if_error_occurs(self):
        with self.storage.client.connect() as conn:
            query = "DELETE FROM objects WHERE resource_name = 'genre';"
            conn.execute(query)

        try:
            with self.storage.client.connect() as conn:
                query = """
                INSERT INTO objects VALUES ('rock-and-roll', 'music', 'genre', NOW(), '{}', FALSE);
                """
                conn.execute(query)
                conn.commit()

                query = """
                INSERT INTO objects VALUES ('jazz', 'music', 'genre', NOW(), '{}', FALSE);
                """
                conn.execute(query)

                raise sqlalchemy.exc.TimeoutError()
        except exceptions.BackendError:
            pass

        with self.storage.client.connect() as conn:
            query = "SELECT COUNT(*) FROM objects WHERE resource_name = 'genre';"
            result = conn.execute(query)
            self.assertEqual(result.fetchone()[0], 1)

    def test_pool_object_is_shared_among_backend_instances(self):
        config = self._get_config()
        storage1 = self.backend.load_from_config(config)
        storage2 = self.backend.load_from_config(config)
        self.assertEqual(id(storage1.client), id(storage2.client))

    def test_warns_if_configured_pool_size_differs_for_same_backend_type(self):
        self.backend.load_from_config(self._get_config())
        settings = {**self.settings, "storage_pool_size": 1}
        msg = "Reuse existing PostgreSQL connection. Parameters storage_* " "will be ignored."
        with mock.patch("kinto.core.storage.postgresql.client." "warnings.warn") as mocked:
            self.backend.load_from_config(self._get_config(settings=settings))
            mocked.assert_any_call(msg)

    def test_list_all_raises_if_missing_on_strange_query(self):
        with self.assertRaises(ValueError):
            self.storage.list_all(
                "some-resource", "some-parent", filters=[Filter("author", MISSING, COMPARISON.HAS)]
            )

    def test_integrity_error_rollsback_transaction(self):
        client = postgresql.create_from_config(
            self._get_config(), prefix="storage_", with_transaction=False
        )
        with self.assertRaises(exceptions.IntegrityError):
            with client.connect() as conn:
                # Make some change in a table.
                conn.execute(
                    """
                INSERT INTO objects
                VALUES ('rock-and-roll', 'music', 'genre', NOW(), '{}', FALSE);
                """
                )
                # Go into a failing integrity constraint.
                query = "INSERT INTO timestamps VALUES ('a', 'b', NOW());"
                conn.execute(query)
                conn.execute(query)
                conn.commit()
                conn.close()

        # Check that change in the above table was rolledback.
        with client.connect() as conn:
            result = conn.execute(
                """
            SELECT FROM objects
             WHERE parent_id = 'music'
               AND resource_name = 'genre';
            """
            )
        self.assertEqual(result.rowcount, 0)

    def test_conflicts_handled_correctly(self):
        config = self._get_config()
        storage = self.backend.load_from_config(config)
        storage.create(resource_name="genre", parent_id="music", obj={"id": "rock-and-roll"})

        def object_not_found(*args, **kwargs):
            raise exceptions.ObjectNotFoundError()

        with mock.patch.object(storage, "get", side_effect=object_not_found):
            with self.assertRaises(exceptions.UnicityError):
                storage.create(
                    resource_name="genre", parent_id="music", obj={"id": "rock-and-roll"}
                )

    def test_supports_null_pool(self):
        settings = {
            **DEFAULT_SETTINGS,
            **self.settings,
            "storage_poolclass": "sqlalchemy.pool.NullPool",
        }
        config = self._get_config(settings=settings)
        self.backend.load_from_config(config)  # does not raise


class PaginatedTest(unittest.TestCase):
    def setUp(self):
        self.storage = mock.Mock()
        self.sample_objects = [
            {"id": "object-01", "flavor": "strawberry"},
            {"id": "object-02", "flavor": "banana"},
            {"id": "object-03", "flavor": "mint"},
            {"id": "object-04", "flavor": "plain"},
            {"id": "object-05", "flavor": "peanut"},
        ]

        def sample_objects_side_effect(*args, **kwargs):
            return self.sample_objects

        self.storage.list_all.side_effect = sample_objects_side_effect

    def test_paginated_passes_sort(self):
        i = paginated(self.storage, sorting=[Sort("id", -1)])
        next(i)  # make the generator do anything
        self.storage.list_all.assert_called_with(
            sorting=[Sort("id", -1)], limit=25, pagination_rules=None
        )

    def test_paginated_passes_batch_size(self):
        i = paginated(self.storage, sorting=[Sort("id", -1)], batch_size=17)
        next(i)  # make the generator do anything
        self.storage.list_all.assert_called_with(
            sorting=[Sort("id", -1)], limit=17, pagination_rules=None
        )

    def test_paginated_yields_objects(self):
        iter = paginated(self.storage, sorting=[Sort("id", -1)])
        assert next(iter) == {"id": "object-01", "flavor": "strawberry"}

    def test_paginated_fetches_next_page(self):
        objects = self.sample_objects
        objects.reverse()

        def list_all_mock(*args, **kwargs):
            this_objects = objects[:3]
            del objects[:3]
            return this_objects

        self.storage.list_all.side_effect = list_all_mock

        list(paginated(self.storage, sorting=[Sort("id", -1)]))
        assert self.storage.list_all.call_args_list == [
            mock.call(sorting=[Sort("id", -1)], limit=25, pagination_rules=None),
            mock.call(
                sorting=[Sort("id", -1)],
                limit=25,
                pagination_rules=[[Filter("id", "object-03", COMPARISON.LT)]],
            ),
            mock.call(
                sorting=[Sort("id", -1)],
                limit=25,
                pagination_rules=[[Filter("id", "object-01", COMPARISON.LT)]],
            ),
        ]



import logging

import pkg_resources
from pyramid.config import Configurator
from pyramid.security import Authenticated, Everyone
from pyramid.settings import asbool

import kinto.core
from kinto.authorization import RouteFactory
from kinto.core import utils

# Module version, as defined in PEP-0396.
__version__ = pkg_resources.get_distribution(__package__).version

# Implemented HTTP API Version
HTTP_API_VERSION = "1.22"

# Main kinto logger
logger = logging.getLogger(__name__)


DEFAULT_SETTINGS = {
    "retry_after_seconds": 3,
    "cache_backend": "kinto.core.cache.memory",
    "permission_backend": "kinto.core.permission.memory",
    "storage_backend": "kinto.core.storage.memory",
    "project_docs": "https://kinto.readthedocs.io/",
    "bucket_create_principals": Authenticated,
    "permissions_read_principals": Everyone,
    "multiauth.authorization_policy": ("kinto.authorization.AuthorizationPolicy"),
    "experimental_collection_schema_validation": False,
    "experimental_permissions_endpoint": False,
    "http_api_version": utils.json_serializer(HTTP_API_VERSION),
    "bucket_id_generator": "kinto.views.NameGenerator",
    "collection_id_generator": "kinto.views.NameGenerator",
    "group_id_generator": "kinto.views.NameGenerator",
    "record_id_generator": "kinto.views.RelaxedUUID",
    "project_name": "kinto",
}


def main(global_config, config=None, **settings):
    if not config:
        config = Configurator(settings=settings, root_factory=RouteFactory)

    config.registry.command = global_config and global_config.get("command", None)

    # Force settings prefix.
    config.add_settings({"kinto.settings_prefix": "kinto"})

    kinto.core.initialize(config, version=__version__, default_settings=DEFAULT_SETTINGS)

    settings = config.get_settings()

    # Expose capability
    schema_enabled = asbool(settings["experimental_collection_schema_validation"])
    if schema_enabled:
        config.add_api_capability(
            "schema",
            description="Validates collection records with JSON schemas.",
            url="https://kinto.readthedocs.io/en/latest/api/1.x/"
            "collections.html#collection-json-schema",
        )

    # Scan Kinto views.
    kwargs = {}

    # Permissions endpoint enabled if permission backend is setup.
    is_admin_enabled = "kinto.plugins.admin" in settings["includes"]
    permissions_endpoint_enabled = (
        is_admin_enabled or asbool(settings["experimental_permissions_endpoint"])
    ) and hasattr(config.registry, "permission")
    if permissions_endpoint_enabled:
        config.add_api_capability(
            "permissions_endpoint",
            description="The permissions endpoint can be used to list all "
            "user objects permissions.",
            url="https://kinto.readthedocs.io/en/latest/configuration/"
            "settings.html#activating-the-permissions-endpoint",
        )
    else:
        kwargs.setdefault("ignore", []).append("kinto.views.permissions")

    config.scan("kinto.views", **kwargs)

    app = config.make_wsgi_app()

    # Install middleware (no-op if disabled)
    return kinto.core.install_middlewares(app, settings)



import base64
import urllib.parse

import colander
import requests
from cornice.validators import colander_validator
from pyramid import httpexceptions

from kinto.core import Service
from kinto.core.errors import ERRORS, raise_invalid
from kinto.core.resource.schema import ErrorResponseSchema
from kinto.core.schema import URL
from kinto.core.utils import random_bytes_hex

from .utils import fetch_openid_config

DEFAULT_STATE_TTL_SECONDS = 3600
DEFAULT_STATE_LENGTH = 32


class RedirectHeadersSchema(colander.MappingSchema):
    """Redirect response headers."""

    location = colander.SchemaNode(colander.String(), name="Location")


class RedirectResponseSchema(colander.MappingSchema):
    """Redirect response schema."""

    headers = RedirectHeadersSchema()


response_schemas = {
    "307": RedirectResponseSchema(description="Successful redirection."),
    "400": ErrorResponseSchema(description="The request is invalid."),
}


def provider_validator(request, **kwargs):
    """
    This validator verifies that the validator in URL (eg. /openid/auth0/login)
    is a configured OpenIDConnect policy.
    """
    provider = request.matchdict["provider"]
    used = request.registry.settings.get("multiauth.policy.%s.use" % provider, "")
    if not used.endswith("OpenIDConnectPolicy"):
        request.errors.add("path", "provider", "Unknow provider %r" % provider)


class LoginQuerystringSchema(colander.MappingSchema):
    """
    Querystring schema for the login endpoint.
    """

    callback = URL()
    scope = colander.SchemaNode(colander.String())
    prompt = colander.SchemaNode(
        colander.String(), validator=colander.Regex("none"), missing=colander.drop
    )


class LoginSchema(colander.MappingSchema):
    querystring = LoginQuerystringSchema()


login = Service(
    name="openid_login", path="/openid/{provider}/login", description="Initiate the OAuth2 login"
)


@login.get(
    schema=LoginSchema(),
    validators=(colander_validator, provider_validator),
    response_schemas=response_schemas,
)
def get_login(request):
    """Initiates to login dance for the specified scopes and callback URI
    using appropriate redirections."""

    # Settings.
    provider = request.matchdict["provider"]
    settings_prefix = "multiauth.policy.%s." % provider
    issuer = request.registry.settings[settings_prefix + "issuer"]
    client_id = request.registry.settings[settings_prefix + "client_id"]
    userid_field = request.registry.settings.get(settings_prefix + "userid_field")
    state_ttl = int(
        request.registry.settings.get(
            settings_prefix + "state_ttl_seconds", DEFAULT_STATE_TTL_SECONDS
        )
    )
    state_length = int(
        request.registry.settings.get(settings_prefix + "state_length", DEFAULT_STATE_LENGTH)
    )

    # Read OpenID configuration (cached by issuer)
    oid_config = fetch_openid_config(issuer)
    auth_endpoint = oid_config["authorization_endpoint"]

    scope = request.GET["scope"]
    callback = request.GET["callback"]
    prompt = request.GET.get("prompt")

    # Check that email scope is requested if userid field is configured as email.
    if userid_field == "email" and "email" not in scope:
        error_details = {
            "name": "scope",
            "description": "Provider %s requires 'email' scope" % provider,
        }
        raise_invalid(request, **error_details)

    # Generate a random string as state.
    # And save it until code is traded.
    state = random_bytes_hex(state_length)
    request.registry.cache.set("openid:state:" + state, callback, ttl=state_ttl)

    # Redirect the client to the Identity Provider that will eventually redirect
    # to the OpenID token endpoint.
    token_uri = request.route_url("openid_token", provider=provider)
    params = dict(
        client_id=client_id, response_type="code", scope=scope, redirect_uri=token_uri, state=state
    )
    if prompt:
        # The 'prompt' parameter is optional.
        params["prompt"] = prompt
    redirect = f"{auth_endpoint}?{urllib.parse.urlencode(params)}"
    raise httpexceptions.HTTPTemporaryRedirect(redirect)


class TokenQuerystringSchema(colander.MappingSchema):
    """
    Querystring schema for the token endpoint.
    """

    code = colander.SchemaNode(colander.String())
    state = colander.SchemaNode(colander.String())


class TokenSchema(colander.MappingSchema):
    querystring = TokenQuerystringSchema()


token = Service(name="openid_token", path="/openid/{provider}/token", description="")


@token.get(schema=TokenSchema(), validators=(colander_validator, provider_validator))
def get_token(request):
    """Trades the specified code and state against access and ID tokens.
    The client is redirected to the original ``callback`` URI with the
    result in querystring."""

    # Settings.
    provider = request.matchdict["provider"]
    settings_prefix = "multiauth.policy.%s." % provider
    issuer = request.registry.settings[settings_prefix + "issuer"]
    client_id = request.registry.settings[settings_prefix + "client_id"]
    client_secret = request.registry.settings[settings_prefix + "client_secret"]

    # Read OpenID configuration (cached by issuer)
    oid_config = fetch_openid_config(issuer)
    token_endpoint = oid_config["token_endpoint"]

    code = request.GET["code"]
    state = request.GET["state"]

    # State can be used only once.
    callback = request.registry.cache.delete("openid:state:" + state)
    if callback is None:
        error_details = {
            "name": "state",
            "description": "Invalid state",
            "errno": ERRORS.INVALID_AUTH_TOKEN.value,
        }
        raise_invalid(request, **error_details)

    # Trade the code for tokens on the Identity Provider.
    # Google Identity requires to specify again redirect_uri.
    redirect_uri = request.route_url("openid_token", provider=provider)
    data = {
        "code": code,
        "client_id": client_id,
        "client_secret": client_secret,
        "redirect_uri": redirect_uri,
        "grant_type": "authorization_code",
    }
    resp = requests.post(token_endpoint, data=data)

    # The IdP response is forwarded to the client in the querystring/location hash.
    # (eg. callback=`http://localhost:3000/#tokens=`)
    token_info = resp.text.encode("utf-8")
    encoded_token = base64.b64encode(token_info)
    redirect = callback + urllib.parse.quote(encoded_token.decode("utf-8"))
    raise httpexceptions.HTTPTemporaryRedirect(redirect)



import contextlib
import logging
import warnings
from collections import defaultdict

import transaction as zope_transaction

from kinto.core.storage import exceptions
from kinto.core.utils import sqlalchemy

logger = logging.getLogger(__name__)

BLACKLISTED_SETTINGS = [
    "backend",
    "max_fetch_size",
    "max_size_bytes",
    "prefix",
    "strict_json",
    "hosts",
]


class PostgreSQLClient:
    def __init__(self, session_factory, commit_manually, invalidate):
        self.session_factory = session_factory
        self.commit_manually = commit_manually
        self.invalidate = invalidate

    @contextlib.contextmanager
    def connect(self, readonly=False, force_commit=False):
        """
        Pulls a connection from the pool when context is entered and
        returns it when context is exited.

        A COMMIT is performed on the current transaction if everything went
        well. Otherwise transaction is ROLLBACK, and everything cleaned up.
        """
        commit_manually = self.commit_manually and not readonly
        session = None
        try:
            # Pull connection from pool.
            session = self.session_factory()
            # Start context
            yield session
            if not readonly and not self.commit_manually:
                # Mark session as dirty.
                self.invalidate(session)
            # Success
            if commit_manually:
                session.commit()
            elif force_commit:
                # Commit like would do a succesful request.
                zope_transaction.commit()

        except sqlalchemy.exc.IntegrityError as e:
            logger.error(e, exc_info=True)
            if commit_manually:  # pragma: no branch
                session.rollback()
            raise exceptions.IntegrityError(original=e) from e
        except sqlalchemy.exc.SQLAlchemyError as e:
            logger.error(e, exc_info=True)
            if session and commit_manually:
                session.rollback()
            raise exceptions.BackendError(original=e) from e
        finally:
            if session and self.commit_manually:
                # Give back to pool if commit done manually.
                session.close()


# Reuse existing client if same URL.
_CLIENTS = defaultdict(dict)


def create_from_config(config, prefix="", with_transaction=True):
    """Create a PostgreSQLClient client using settings in the provided config.
    """
    if sqlalchemy is None:
        message = (
            "PostgreSQL SQLAlchemy dependency missing. "
            "Refer to installation section in documentation."
        )
        raise ImportWarning(message)

    from zope.sqlalchemy import register, invalidate
    from sqlalchemy.orm import sessionmaker, scoped_session

    settings = {**config.get_settings()}

    # Custom Kinto settings, unsupported by SQLAlchemy.
    blacklist = [prefix + setting for setting in BLACKLISTED_SETTINGS]
    filtered_settings = {k: v for k, v in settings.items() if k not in blacklist}
    transaction_per_request = with_transaction and filtered_settings.pop(
        "transaction_per_request", False
    )
    url = filtered_settings[prefix + "url"]
    existing_client = _CLIENTS[transaction_per_request].get(url)
    if existing_client:
        msg = "Reuse existing PostgreSQL connection. " f"Parameters {prefix}* will be ignored."
        warnings.warn(msg)
        return existing_client

    # Initialize SQLAlchemy engine from filtered_settings.
    poolclass_key = prefix + "poolclass"
    filtered_settings.setdefault(
        poolclass_key, ("kinto.core.storage.postgresql." "pool.QueuePoolWithMaxBacklog")
    )
    filtered_settings[poolclass_key] = config.maybe_dotted(filtered_settings[poolclass_key])
    engine = sqlalchemy.engine_from_config(filtered_settings, prefix=prefix, url=url)

    # Initialize thread-safe session factory.
    session_factory = scoped_session(sessionmaker(bind=engine))
    if transaction_per_request:
        # Plug with Pyramid transaction manager
        register(session_factory)

    # Store one client per URI.
    commit_manually = not transaction_per_request
    client = PostgreSQLClient(session_factory, commit_manually, invalidate)
    _CLIENTS[transaction_per_request][url] = client
    return client



import os
import unittest
from unittest import mock

import colander
from pyramid import httpexceptions
from pyramid import request as pyramid_request
from pyramid import testing

from kinto.core import DEFAULT_SETTINGS, includeme
from kinto.core.testing import DummyRequest
from kinto.core.utils import (
    build_request,
    current_service,
    dict_merge,
    dict_subset,
    find_nested_value,
    follow_subrequest,
    hmac_digest,
    instance_uri_registry,
    native_value,
    prefixed_principals,
    random_bytes_hex,
    read_env,
    recursive_update_dict,
    strip_whitespace,
)


def build_real_request(wsgi_environ):
    """Build a Pyramid request, as if it was instantiated by Pyramid.
    """
    config = testing.setUp(settings=DEFAULT_SETTINGS)
    includeme(config)
    request = pyramid_request.Request(wsgi_environ)
    request.registry = config.registry
    return request


class NativeValueTest(unittest.TestCase):
    def test_simple_string(self):
        self.assertEqual(native_value("value"), "value")

    def test_defined_string(self):
        self.assertEqual(native_value('"value"'), "value")

    def test_null_value(self):
        self.assertEqual(native_value("null"), None)

    def test_defined_null_as_text_value(self):
        self.assertEqual(native_value('"null"'), "null")

    def test_integer(self):
        self.assertEqual(native_value("7"), 7)

    def test_defined_integer_as_text_value(self):
        self.assertEqual(native_value('"7"'), "7")

    def test_defined_simple_quote_string_as_text_value(self):
        self.assertEqual(native_value("'7'"), "'7'")

    def test_zero_and_one_coerce_to_integers(self):
        self.assertEqual(native_value("1"), 1)
        self.assertEqual(native_value("0"), 0)

    def test_float(self):
        self.assertEqual(native_value("3.14"), 3.14)

    def test_true_values(self):
        true_strings = ["true"]
        true_values = [native_value(s) for s in true_strings]
        self.assertTrue(all(true_values))

    def test_false_values(self):
        false_strings = ["false"]
        false_values = [native_value(s) for s in false_strings]
        self.assertFalse(any(false_values))

    def test_non_string_values(self):
        self.assertEqual(native_value(7), 7)
        self.assertEqual(native_value(True), True)

    def test_bad_string_values(self):
        self.assertEqual(native_value("\u0000"), "\x00")

    def test_infinite_strings(self):
        self.assertEqual(native_value("803E5708"), "803E5708")
        self.assertEqual(native_value("Infinity"), "Infinity")


class StripWhitespaceTest(unittest.TestCase):
    def test_removes_all_kinds_of_spaces(self):
        value = " \t teaser \n \r"
        self.assertEqual(strip_whitespace(value), "teaser")

    def test_does_remove_middle_spaces(self):
        self.assertEqual(strip_whitespace("a b c"), "a b c")

    def test_idempotent_for_null_values(self):
        self.assertEqual(strip_whitespace(colander.null), colander.null)


class CryptographicRandomBytesTest(unittest.TestCase):
    def test_return_hex_string(self):
        value = random_bytes_hex(16)
        try:
            int(value, 16)
        except ValueError:
            self.fail("{} is not an hexadecimal value.".format(value))

    def test_return_right_length_string(self):
        for x in range(2, 4):
            value = random_bytes_hex(x)
            self.assertEqual(len(value), x * 2)

    def test_return_text_string(self):
        value = random_bytes_hex(16)
        self.assertIsInstance(value, str)


class HmacDigestTest(unittest.TestCase):
    def test_supports_secret_as_text(self):
        value = hmac_digest("blah", "input data")
        self.assertTrue(value.startswith("d4f5c51db246c7faeb42240545b47274b6"))

    def test_supports_secret_as_bytes(self):
        value = hmac_digest(b"blah", "input data")
        self.assertTrue(value.startswith("d4f5c51db246c7faeb42240545b47274b6"))


class ReadEnvironmentTest(unittest.TestCase):
    def test_return_passed_value_if_not_defined_in_env(self):
        self.assertEqual(read_env("missing", 12), 12)

    def test_return_env_value_if_defined_in_env(self):
        os.environ.setdefault("KINTO_CONF", "abc")
        self.assertEqual(read_env("KINTO_CONF", 12), "abc")

    def test_return_env_name_as_uppercase(self):
        os.environ.setdefault("KINTO_NAME", "abc")
        self.assertEqual(read_env("kinto.name", 12), "abc")

    def test_return_env_value_is_coerced_to_python(self):
        os.environ.setdefault("KINTO_CONF_NAME", "3.14")
        self.assertEqual(read_env("kinto-conf.name", 12), 3.14)

    def test_default_value_is_not_coerced_to_python(self):
        self.assertEqual(read_env("kinto-conf.name", "12"), "12")


class CurrentServiceTest(unittest.TestCase):
    def test_current_service_returns_the_service_for_existing_patterns(self):
        request = DummyRequest()
        request.matched_route.pattern = "/buckets"
        request.registry.cornice_services = {"/buckets": mock.sentinel.service}

        self.assertEqual(current_service(request), mock.sentinel.service)

    def test_current_service_returns_none_for_unexisting_patterns(self):
        request = DummyRequest()
        request.matched_route.pattern = "/unexisting"
        request.registry.cornice_services = {}

        self.assertEqual(current_service(request), None)


class PrefixedPrincipalsTest(unittest.TestCase):
    def test_removes_unprefixed_from_principals(self):
        request = DummyRequest()
        request.effective_principals = ["foo", "system.Authenticated"]
        request.prefixed_userid = "basic:foo"
        self.assertEqual(prefixed_principals(request), ["basic:foo", "system.Authenticated"])

    def test_works_if_userid_is_not_in_principals(self):
        request = DummyRequest()
        request.effective_principals = ["basic:foo", "system.Authenticated"]
        request.prefixed_userid = "basic:foo"
        self.assertEqual(prefixed_principals(request), ["basic:foo", "system.Authenticated"])


class BuildRequestTest(unittest.TestCase):
    def test_built_request_has_kinto_core_custom_methods(self):
        original = build_real_request({"PATH_INFO": "/foo"})
        request = build_request(original, {"path": "bar"})
        self.assertTrue(hasattr(request, "current_service"))


class FollowSubrequestTest(unittest.TestCase):
    def test_parent_and_bound_data_are_preserved(self):
        request = DummyRequest()
        request.invoke_subrequest.side_effect = (httpexceptions.HTTPTemporaryRedirect, None)
        subrequest = DummyRequest()
        subrequest.parent = mock.sentinel.parent
        subrequest.bound_data = mock.sentinel.bound_data
        _, redirected = follow_subrequest(request, subrequest)
        self.assertEqual(subrequest.parent, redirected.parent)
        self.assertEqual(subrequest.bound_data, redirected.bound_data)


class DictSubsetTest(unittest.TestCase):
    def test_extract_by_keys(self):
        obtained = dict_subset(dict(a=1, b=2), ["b"])
        expected = dict(b=2)
        self.assertEqual(obtained, expected)

    def test_is_noop_if_no_keys(self):
        obtained = dict_subset(dict(a=1, b=2), [])
        expected = dict()
        self.assertEqual(obtained, expected)

    def test_ignores_unknown_keys(self):
        obtained = dict_subset(dict(a=1, b=2), ["a", "a.b", "d.b", "c"])
        expected = dict(a=1)
        self.assertEqual(obtained, expected)

    def test_ignores_duplicated_keys(self):
        obtained = dict_subset(dict(a=1, b=2), ["a", "a"])
        expected = dict(a=1)
        self.assertEqual(obtained, expected)

    def test_can_filter_subobjects(self):
        obtained = dict_subset(dict(a=1, b=dict(c=2, d=3)), ["a", "b.c"])
        expected = dict(a=1, b=dict(c=2))
        self.assertEqual(obtained, expected)

    def test_can_filter_subobjects_keys(self):
        input = dict(a=1, b=dict(c=2, d=3, e=4))
        obtained = dict_subset(input, ["a", "b.d", "b.e"])
        expected = dict(a=1, b=dict(d=3, e=4))
        self.assertEqual(obtained, expected)

    def test_can_filter_subobjects_recursively(self):
        input = dict(b=dict(c=2, d=dict(e=4, f=5, g=6)))
        obtained = dict_subset(input, ["b.d.e", "b.d.f"])
        expected = dict(b=dict(d=dict(e=4, f=5)))
        self.assertEqual(obtained, expected)

    def test_ignores_if_subobject_is_not_dict(self):
        input = dict(a=1, b=dict(c=2, d=3))
        obtained = dict_subset(input, ["a", "b.c.d", "b.d"])
        expected = dict(a=1, b=dict(c=2, d=3))
        self.assertEqual(obtained, expected)


class DictMergeTest(unittest.TestCase):
    def test_merge(self):
        obtained = dict_merge(dict(a=1, b=dict(c=2)), dict(b=dict(d=4)))
        expected = dict(a=1, b=dict(c=2, d=4))
        self.assertEqual(obtained, expected)


class FindNestedValueTest(unittest.TestCase):
    def test_find_flat_value(self):
        obj = {"a": 42}
        obtained = find_nested_value(obj, "a")
        self.assertEqual(obtained, 42)

    def test_find_nested_value(self):
        obj = {"a": {"b": 42}}
        obtained = find_nested_value(obj, "a.b")
        self.assertEqual(obtained, 42)

    def test_find_deeply_nested_value(self):
        obj = {"a": {"b": {"c": 42}}}
        obtained = find_nested_value(obj, "a.b.c")
        self.assertEqual(obtained, 42)

    def test_find_dotted_path_value(self):
        obj = {"a.b": 42}
        obtained = find_nested_value(obj, "a.b")
        self.assertEqual(obtained, 42)

    def test_find_nested_dotted_path_value(self):
        obj = {"a": {"b.c": 42}}
        obtained = find_nested_value(obj, "a.b.c")
        self.assertEqual(obtained, 42)

        obj = {"a.b": {"c.d": 42}}
        obtained = find_nested_value(obj, "a.b.c.d")
        self.assertEqual(obtained, 42)

        obj = {"a": {"b": {"a": {"b": 42}}}}
        obtained = find_nested_value(obj, "a.b.a.b")
        self.assertEqual(obtained, 42)

    def test_find_disambiguated_dotted_path_values(self):
        # XXX: To be honest, this is a really scary use case. Probably a
        # limitation of the dotted path notation we may want to document.
        # At least this test acts as documentation for now.
        obj = {"a": {"b": 0}, "a.b": 42}
        obtained = find_nested_value(obj, "a.b")
        self.assertEqual(obtained, 42)

    def test_unmatched_path_returns_none(self):
        obj = {"a": 42}
        self.assertIsNone(find_nested_value(obj, "x"))
        self.assertIsNone(find_nested_value(obj, "a.b"))
        self.assertIsNone(find_nested_value(obj, "x.a"))

    def test_fallback_default_value(self):
        obj = {"a": {"c": 42}}
        self.assertEqual(find_nested_value(obj, "x", 1337), 1337)
        self.assertEqual(find_nested_value(obj, "a.b", 1337), 1337)
        self.assertEqual(find_nested_value(obj, "x.a", 1337), 1337)
        self.assertEqual(find_nested_value(obj, "a.c.d", 1337), 1337)


class RecursiveUpdateDictTest(unittest.TestCase):
    def test_merge(self):
        a = {}
        recursive_update_dict(a, {"b": {"c": 1}, "d": 2})
        self.assertEqual(a["b"]["c"], 1)
        self.assertEqual(a["d"], 2)

    def test_merge_non_dict(self):
        a = {}
        recursive_update_dict(a, 1)
        self.assertEqual(a, {})


class InstanceURIRegistryTest(unittest.TestCase):
    @mock.patch("kinto.core.utils.instance_uri")
    def test_instance_uri_registry_calls_instance_uri(self, instance_uri):
        registry = mock.Mock()
        instance_uri_registry(registry, "object", a=1)
        self.assertEqual(len(instance_uri.call_args_list), 1)
        (args, kwargs) = instance_uri.call_args_list[0]
        self.assertEqual(len(args), 2)

        self.assertEqual(args[0].registry, registry)
        self.assertEqual(args[1], "object")

        self.assertEqual(kwargs, {"a": 1})



import json
import unittest

from kinto.core.errors import send_alert
from kinto.core.testing import DummyRequest


class SendAlertTest(unittest.TestCase):
    def verify_alert_header(self, request, expected):
        self.assertIn("Alert", request.response.headers)
        alert = request.response.headers["Alert"]
        self.assertDictEqual(json.loads(alert), expected)

    def test_send_alert_default_to_project_url(self):
        request = DummyRequest()
        request.registry.settings["project_docs"] = "docs_url"
        send_alert(request, "Message")
        self.verify_alert_header(
            request, {"code": "soft-eol", "message": "Message", "url": "docs_url"}
        )

    def test_send_alert_url_can_be_specified(self):
        request = DummyRequest()
        send_alert(request, "Message", "error_url")
        self.verify_alert_header(
            request, {"code": "soft-eol", "message": "Message", "url": "error_url"}
        )

    def test_send_alert_code_can_be_specified(self):
        request = DummyRequest()
        request.registry.settings["project_docs"] = "docs_url"
        send_alert(request, "Message", code="hard-eol")
        self.verify_alert_header(
            request, {"code": "hard-eol", "message": "Message", "url": "docs_url"}
        )



from kinto.core.events import ResourceChanged

from .listener import on_resource_changed


def includeme(config):
    config.add_api_capability(
        "quotas",
        description="Quotas Management on buckets " "and collections.",
        url="https://kinto.readthedocs.io",
    )

    # If StatsD is enabled, monitor execution time of listener.
    listener = on_resource_changed
    if config.registry.statsd:
        key = "plugins.quotas"
        listener = config.registry.statsd.timer(key)(on_resource_changed)

    # Listen to every resources (except history)
    config.add_subscriber(
        listener, ResourceChanged, for_resources=("bucket", "group", "collection", "record")
    )



import random
import string

from pyramid.httpexceptions import HTTPNotFound

from kinto.core.errors import ERRORS, http_error
from kinto.core.storage import exceptions, generators


class NameGenerator(generators.Generator):
    def __call__(self):
        alpha_num = string.ascii_letters + string.digits
        alphabet = alpha_num + "-_"
        letters = [random.SystemRandom().choice(alpha_num)]
        letters += [random.SystemRandom().choice(alphabet) for x in range(7)]

        return "".join(letters)


class RelaxedUUID(generators.UUID4):
    """A generator that generates UUIDs but accepts any string.
    """

    regexp = generators.Generator.regexp


def object_exists_or_404(request, resource_name, object_id, parent_id=""):
    storage = request.registry.storage
    try:
        return storage.get(resource_name=resource_name, parent_id=parent_id, object_id=object_id)
    except exceptions.ObjectNotFoundError:
        # XXX: We gave up putting details about parent id here (See #53).
        details = {"id": object_id, "resource_name": resource_name}
        response = http_error(HTTPNotFound(), errno=ERRORS.MISSING_RESOURCE, details=details)
        raise response



import warnings

import colander

from kinto.core.errors import ErrorSchema
from kinto.core.schema import (
    URL,
    Any,
    FieldList,
    HeaderField,
    HeaderQuotedInteger,
    QueryField,
    TimeStamp,
)
from kinto.core.utils import native_value

POSTGRESQL_MAX_INTEGER_VALUE = 2 ** 63

positive_big_integer = colander.Range(min=0, max=POSTGRESQL_MAX_INTEGER_VALUE)


class TimeStamp(TimeStamp):
    """This schema is deprecated, you shoud use `kinto.core.schema.TimeStamp` instead."""

    def __init__(self, *args, **kwargs):
        message = (
            "`kinto.core.resource.schema.TimeStamp` is deprecated, "
            "use `kinto.core.schema.TimeStamp` instead."
        )
        warnings.warn(message, DeprecationWarning)
        super().__init__(*args, **kwargs)


class URL(URL):
    """This schema is deprecated, you shoud use `kinto.core.schema.URL` instead."""

    def __init__(self, *args, **kwargs):
        message = (
            "`kinto.core.resource.schema.URL` is deprecated, "
            "use `kinto.core.schema.URL` instead."
        )
        warnings.warn(message, DeprecationWarning)
        super().__init__(*args, **kwargs)


# Resource related schemas


class ResourceSchema(colander.MappingSchema):
    """Base resource schema, with *Cliquet* specific built-in options."""

    class Options:
        """
        Resource schema options.

        This is meant to be overriden for changing values:

        .. code-block:: python

            class Product(ResourceSchema):
                reference = colander.SchemaNode(colander.String())

                class Options:
                    readonly_fields = ('reference',)
        """

        readonly_fields = tuple()
        """Fields that cannot be updated. Values for fields will have to be
        provided either during object creation, through default values using
        ``missing`` attribute or implementing a custom logic in
        :meth:`kinto.core.resource.Resource.process_object`.
        """

        preserve_unknown = True
        """Define if unknown fields should be preserved or not.

        The resource is schema-less by default. In other words, any field name
        will be accepted on objects. Set this to ``False`` in order to limit
        the accepted fields to the ones defined in the schema.
        """

    @classmethod
    def get_option(cls, attr):
        default_value = getattr(ResourceSchema.Options, attr)
        return getattr(cls.Options, attr, default_value)

    @classmethod
    def is_readonly(cls, field):
        """Return True if specified field name is read-only.

        :param str field: the field name in the schema
        :returns: ``True`` if the specified field is read-only,
            ``False`` otherwise.
        :rtype: bool
        """
        return field in cls.get_option("readonly_fields")

    def schema_type(self):
        if self.get_option("preserve_unknown") is True:
            unknown = "preserve"
        else:
            unknown = "ignore"
        return colander.Mapping(unknown=unknown)


class PermissionsSchema(colander.SchemaNode):
    """A permission mapping defines ACEs.

    It has permission names as keys and principals as values.

    ::

        {
            "write": ["fxa:af3e077eb9f5444a949ad65aa86e82ff"],
            "groups:create": ["fxa:70a9335eecfe440fa445ba752a750f3d"]
        }

    """

    def __init__(self, *args, **kwargs):
        self.known_perms = kwargs.pop("permissions", tuple())
        super().__init__(*args, **kwargs)

        for perm in self.known_perms:
            self[perm] = self._get_node_principals(perm)

    def schema_type(self):
        if self.known_perms:
            return colander.Mapping(unknown="raise")
        else:
            return colander.Mapping(unknown="preserve")

    def deserialize(self, cstruct=colander.null):
        # If permissions are not a mapping (e.g null or invalid), try deserializing
        if not isinstance(cstruct, dict):
            return super().deserialize(cstruct)

        # If using application/merge-patch+json we need to allow null values as they
        # represent removing a key.
        cstruct, removed_keys = self._preprocess_null_perms(cstruct)

        # If permissions are listed, check fields and produce fancy error messages
        if self.known_perms:
            for perm in cstruct:
                colander.OneOf(choices=self.known_perms)(self, perm)
            permissions = super().deserialize(cstruct)

        # Else deserialize the fields that are not on the schema
        else:
            permissions = {}
            perm_schema = colander.SequenceSchema(colander.SchemaNode(colander.String()))
            for perm, principals in cstruct.items():
                permissions[perm] = perm_schema.deserialize(principals)

        return self._postprocess_null_perms(permissions, removed_keys)

    def _get_node_principals(self, perm):
        principal = colander.SchemaNode(colander.String())
        return colander.SchemaNode(
            colander.Sequence(), principal, name=perm, missing=colander.drop
        )

    @staticmethod
    def _preprocess_null_perms(cstruct):
        keys = {k for k, v in cstruct.items() if v is None}
        cleaned = {k: v for k, v in cstruct.items() if v is not None}
        return cleaned, keys

    @staticmethod
    def _postprocess_null_perms(validated, keys):
        validated.update({k: None for k in keys})
        return validated


# Header schemas


class HeaderSchema(colander.MappingSchema):
    """Base schema used for validating and deserializing request headers. """

    missing = colander.drop

    if_match = HeaderQuotedInteger(name="If-Match")
    if_none_match = HeaderQuotedInteger(name="If-None-Match")

    @staticmethod
    def schema_type():
        return colander.Mapping(unknown="preserve")


class PatchHeaderSchema(HeaderSchema):
    """Header schema used with PATCH requests."""

    def response_behavior_validator():
        return colander.OneOf(["full", "light", "diff"])

    response_behaviour = HeaderField(
        colander.String(), name="Response-Behavior", validator=response_behavior_validator()
    )


# Querystring schemas


class QuerySchema(colander.MappingSchema):
    """
    Schema used for validating and deserializing querystrings. It will include
    and try to guess the type of unknown fields (field filters) on deserialization.
    """

    missing = colander.drop

    @staticmethod
    def schema_type():
        return colander.Mapping(unknown="ignore")

    def deserialize(self, cstruct=colander.null):
        """
        Deserialize and validate the QuerySchema fields and try to deserialize and
        get the native value of additional filds (field filters) that may be present
        on the cstruct.

        e.g:: ?exclude_id=a,b&deleted=true -> {'exclude_id': ['a', 'b'], deleted: True}
        """
        values = {}

        schema_values = super().deserialize(cstruct)
        if schema_values is colander.drop:
            return schema_values

        # Deserialize querystring field filters (see docstring e.g)
        for k, v in cstruct.items():
            # Deserialize lists used on contains_ and contains_any_ filters
            if k.startswith("contains_"):
                as_list = native_value(v)

                if not isinstance(as_list, list):
                    values[k] = [as_list]
                else:
                    values[k] = as_list

            # Deserialize lists used on in_ and exclude_ filters
            elif k.startswith("in_") or k.startswith("exclude_"):
                as_list = FieldList().deserialize(v)
                values[k] = [native_value(v) for v in as_list]
            else:
                values[k] = native_value(v)

        values.update(schema_values)
        return values


class PluralQuerySchema(QuerySchema):
    """Querystring schema used with plural endpoints."""

    _limit = QueryField(colander.Integer(), validator=positive_big_integer)
    _sort = FieldList()
    _token = QueryField(colander.String())
    _since = QueryField(colander.Integer(), validator=positive_big_integer)
    _to = QueryField(colander.Integer(), validator=positive_big_integer)
    _before = QueryField(colander.Integer(), validator=positive_big_integer)
    id = QueryField(colander.String())
    last_modified = QueryField(colander.Integer(), validator=positive_big_integer)


class ObjectGetQuerySchema(QuerySchema):
    """Querystring schema for GET object requests."""

    _fields = FieldList()


class PluralGetQuerySchema(PluralQuerySchema):
    """Querystring schema for GET plural endpoints requests."""

    _fields = FieldList()


# Body Schemas


class ObjectSchema(colander.MappingSchema):
    @colander.deferred
    def data(node, kwargs):
        data = kwargs.get("data")
        if data:
            # Check if empty object is allowed.
            # (e.g every schema fields have defaults)
            try:
                data.deserialize({})
            except colander.Invalid:
                pass
            else:
                data.default = {}
                data.missing = colander.drop
        return data

    @colander.deferred
    def permissions(node, kwargs):
        def get_perms(node, kwargs):
            return kwargs.get("permissions")

        # Set if node is provided, else keep deferred. This allows binding the body
        # on Resource first and bind permissions later.
        # XXX: probably not necessary now that UserResource is gone.
        return get_perms(node, kwargs) or colander.deferred(get_perms)

    @staticmethod
    def schema_type():
        return colander.Mapping(unknown="raise")


class JsonPatchOperationSchema(colander.MappingSchema):
    """Single JSON Patch Operation."""

    def op_validator():
        op_values = ["test", "add", "remove", "replace", "move", "copy"]
        return colander.OneOf(op_values)

    def path_validator():
        return colander.Regex("(/\\w*)+")

    op = colander.SchemaNode(colander.String(), validator=op_validator())
    path = colander.SchemaNode(colander.String(), validator=path_validator())
    from_ = colander.SchemaNode(
        colander.String(), name="from", validator=path_validator(), missing=colander.drop
    )
    value = colander.SchemaNode(Any(), missing=colander.drop)

    @staticmethod
    def schema_type():
        return colander.Mapping(unknown="raise")


class JsonPatchBodySchema(colander.SequenceSchema):
    """Body used with JSON Patch (application/json-patch+json) as in RFC 6902."""

    operations = JsonPatchOperationSchema(missing=colander.drop)


# Request schemas


class RequestSchema(colander.MappingSchema):
    """Base schema for kinto requests."""

    @colander.deferred
    def header(node, kwargs):
        return kwargs.get("header")

    @colander.deferred
    def querystring(node, kwargs):
        return kwargs.get("querystring")

    def after_bind(self, node, kw):
        # Set default bindings
        if not self.get("header"):
            self["header"] = HeaderSchema()
        if not self.get("querystring"):
            self["querystring"] = QuerySchema()


class PayloadRequestSchema(RequestSchema):
    """Base schema for methods that use a JSON request body."""

    @colander.deferred
    def body(node, kwargs):
        def get_body(node, kwargs):
            return kwargs.get("body")

        # Set if node is provided, else keep deferred (and allow bindind later)
        return get_body(node, kwargs) or colander.deferred(get_body)


class JsonPatchRequestSchema(RequestSchema):
    """JSON Patch (application/json-patch+json) request schema."""

    body = JsonPatchBodySchema()
    querystring = QuerySchema()
    header = PatchHeaderSchema()


# Response schemas


class ResponseHeaderSchema(colander.MappingSchema):
    """Kinto API custom response headers."""

    etag = HeaderQuotedInteger(name="Etag")
    last_modified = colander.SchemaNode(colander.String(), name="Last-Modified")


class ErrorResponseSchema(colander.MappingSchema):
    """Response schema used on 4xx and 5xx errors."""

    body = ErrorSchema()


class NotModifiedResponseSchema(colander.MappingSchema):
    """Response schema used on 304 Not Modified responses."""

    header = ResponseHeaderSchema()


class ObjectResponseSchema(colander.MappingSchema):
    """Response schema used with sigle resource endpoints."""

    header = ResponseHeaderSchema()

    @colander.deferred
    def body(node, kwargs):
        return kwargs.get("object")


class PluralResponseSchema(colander.MappingSchema):
    """Response schema used with plural endpoints."""

    header = ResponseHeaderSchema()

    @colander.deferred
    def body(node, kwargs):
        resource = kwargs.get("object")["data"]
        datalist = colander.MappingSchema()
        datalist["data"] = colander.SequenceSchema(resource, missing=[])
        return datalist


class ResourceReponses:
    """Class that wraps and handles Resource responses."""

    default_schemas = {
        "400": ErrorResponseSchema(description="The request is invalid."),
        "401": ErrorResponseSchema(description="The request is missing authentication headers."),
        "403": ErrorResponseSchema(
            description=(
                "The user is not allowed to perform the operation, "
                "or the resource is not accessible."
            )
        ),
        "406": ErrorResponseSchema(
            description="The client doesn't accept supported responses Content-Type."
        ),
        "412": ErrorResponseSchema(
            description="Object was changed or deleted since value in `If-Match` header."
        ),
        "default": ErrorResponseSchema(description="Unexpected error."),
    }
    default_object_schemas = {"200": ObjectResponseSchema(description="Return the target object.")}
    default_plural_schemas = {
        "200": PluralResponseSchema(description="Return a list of matching objects.")
    }
    default_get_schemas = {
        "304": NotModifiedResponseSchema(
            description="Reponse has not changed since value in If-None-Match header"
        )
    }
    default_post_schemas = {
        "200": ObjectResponseSchema(description="Return an existing object."),
        "201": ObjectResponseSchema(description="Return a created object."),
        "415": ErrorResponseSchema(
            description="The client request was not sent with a correct Content-Type."
        ),
    }
    default_put_schemas = {
        "201": ObjectResponseSchema(description="Return created object."),
        "415": ErrorResponseSchema(
            description="The client request was not sent with a correct Content-Type."
        ),
    }
    default_patch_schemas = {
        "415": ErrorResponseSchema(
            description="The client request was not sent with a correct Content-Type."
        )
    }
    default_delete_schemas = {}
    object_get_schemas = {
        "404": ErrorResponseSchema(description="The object does not exist or was deleted.")
    }
    object_patch_schemas = {
        "404": ErrorResponseSchema(description="The object does not exist or was deleted.")
    }
    object_delete_schemas = {
        "404": ErrorResponseSchema(description="The object does not exist or was already deleted.")
    }

    def get_and_bind(self, endpoint_type, method, **kwargs):
        """Wrap resource colander response schemas for an endpoint and return a dict
        of status codes mapping cloned and binded responses."""

        responses = self.default_schemas.copy()
        type_responses = getattr(self, f"default_{endpoint_type}_schemas")
        responses.update(**type_responses)

        verb_responses = f"default_{method.lower()}_schemas"
        method_args = getattr(self, verb_responses, {})
        responses.update(**method_args)

        method_responses = f"{endpoint_type}_{method.lower()}_schemas"
        endpoint_args = getattr(self, method_responses, {})
        responses.update(**endpoint_args)

        # Bind and clone schemas into a new dict
        bound = {code: resp.bind(**kwargs) for code, resp in responses.items()}

        return bound



import logging
from concurrent.futures import ThreadPoolExecutor, wait

import colander
import transaction
from pyramid.security import NO_PERMISSION_REQUIRED

from kinto.core import Service

logger = logging.getLogger(__name__)


heartbeat = Service(name="heartbeat", path="/__heartbeat__", description="Server health")


class HeartbeatResponseSchema(colander.MappingSchema):
    body = colander.SchemaNode(colander.Mapping(unknown="preserve"))


heartbeat_responses = {
    "200": HeartbeatResponseSchema(description="Server is working properly."),
    "503": HeartbeatResponseSchema(description="One or more subsystems failing."),
}


@heartbeat.get(
    permission=NO_PERMISSION_REQUIRED,
    tags=["Utilities"],
    operation_id="__heartbeat__",
    response_schemas=heartbeat_responses,
)
def get_heartbeat(request):
    """Return information about server health."""
    status = {}

    def heartbeat_check(name, func):
        status[name] = False
        status[name] = func(request)
        # Since the heartbeat checks run concurrently, their transactions
        # overlap and might end in shared lock errors. By aborting here
        # we clean-up the state on each heartbeat call instead of once at the
        # end of the request. See bug Kinto/kinto#804
        transaction.abort()

    # Start executing heartbeats concurrently.
    heartbeats = request.registry.heartbeats
    pool = ThreadPoolExecutor(max_workers=max(1, len(heartbeats.keys())))
    futures = []
    for name, func in heartbeats.items():
        future = pool.submit(heartbeat_check, name, func)
        future.__heartbeat_name = name  # For logging purposes.
        futures.append(future)

    # Wait for the results, with timeout.
    seconds = float(request.registry.settings["heartbeat_timeout_seconds"])
    done, not_done = wait(futures, timeout=seconds)

    # A heartbeat is supposed to return True or False, and never raise.
    # Just in case, go though results to spot any potential exception.
    for future in done:
        exc = future.exception()
        if exc is not None:
            logger.error(f"'{future.__heartbeat_name}' heartbeat failed.")
            logger.error(exc)

    # Log timed-out heartbeats.
    for future in not_done:
        name = future.__heartbeat_name
        error_msg = f"'{name}' heartbeat has exceeded timeout of {seconds} seconds."
        logger.error(error_msg)

    # If any has failed, return a 503 error response.
    has_error = not all([v or v is None for v in status.values()])
    if has_error:
        request.response.status = 503

    return status


class LbHeartbeatResponseSchema(colander.MappingSchema):
    body = colander.SchemaNode(colander.Mapping())


lbheartbeat_responses = {
    "200": LbHeartbeatResponseSchema(description="Returned if server is reachable.")
}


lbheartbeat = Service(name="lbheartbeat", path="/__lbheartbeat__", description="Web head health")


@lbheartbeat.get(
    permission=NO_PERMISSION_REQUIRED,
    tags=["Utilities"],
    operation_id="__lbheartbeat__",
    response_schemas=lbheartbeat_responses,
)
def get_lbheartbeat(request):
    """Return successful healthy response.

    If the load-balancer tries to access this URL and fails, this means the
    Web head is not operational and should be dropped.
    """
    status = {}
    return status



import unittest
from unittest import mock

from kinto.core import resource
from kinto.core.authorization import RouteFactory
from kinto.core.cache import memory as cache_memory
from kinto.core.storage import memory as storage_memory
from kinto.core.testing import DummyRequest


class StrictSchema(resource.ResourceSchema):
    class Options:
        preserve_unknown = False


class BaseTest(unittest.TestCase):
    resource_class = resource.Resource

    def setUp(self):
        self.storage = storage_memory.Storage()
        self.cache = cache_memory.Cache(cache_prefix="", cache_max_size_bytes=1000)
        self.resource = self.resource_class(request=self.get_request(), context=self.get_context())
        self.resource.schema = StrictSchema
        self.model = self.resource.model
        self.patch_known_field = mock.patch.object(self.resource, "is_known_field")
        self.validated = {"body": {}, "header": {}, "querystring": {}}
        self.resource.request.validated = self.validated

    def tearDown(self):
        mock.patch.stopall()

    def get_request(self, resource_name=""):
        request = DummyRequest(method="GET")
        request.current_resource_name = resource_name
        request.registry.cache = self.cache
        request.registry.storage = self.storage
        return request

    def get_context(self):
        return RouteFactory(self.get_request())

    @property
    def last_response(self):
        return self.resource.request.response



from unittest import mock

import webtest
from pyramid.config import Configurator
from pyramid.events import NewRequest
from pyramid.exceptions import ConfigurationError

import kinto.core
from kinto.core import initialization
from kinto.core.testing import unittest


class InitializationTest(unittest.TestCase):
    def test_fails_if_no_version_is_specified(self):
        config = Configurator()
        self.assertRaises(ConfigurationError, kinto.core.initialize, config)

    def test_uses_the_version_for_prefix(self):
        config = Configurator()
        kinto.core.initialize(config, "0.0.1", "name")
        self.assertEqual(config.route_prefix, "v0")

    def test_set_the_project_version_if_specified(self):
        config = Configurator()
        kinto.core.initialize(config, "0.0.1", "name")
        self.assertEqual(config.registry.settings["project_version"], "0.0.1")

    def test_project_version_uses_setting_if_specified(self):
        config = Configurator(settings={"name.project_version": "1.0.0"})
        kinto.core.initialize(config, "0.0.1", "name")
        self.assertEqual(config.registry.settings["project_version"], "1.0.0")

    def test_http_api_version_relies_on_project_version_by_default(self):
        config = Configurator()
        kinto.core.initialize(config, "0.1.0", "name")
        self.assertEqual(config.registry.settings["http_api_version"], "0.1")

    def test_http_api_version_uses_setting_if_specified(self):
        config = Configurator(settings={"name.http_api_version": "1.3"})
        kinto.core.initialize(config, "0.0.1", "name")
        self.assertEqual(config.registry.settings["http_api_version"], "1.3")

    def test_warns_if_settings_prefix_is_empty(self):
        config = Configurator(settings={"kinto.settings_prefix": ""})
        with mock.patch("kinto.core.initialization.warnings.warn") as mocked:
            kinto.core.initialize(config, "0.0.1")
            error_msg = "No value specified for `settings_prefix`"
            mocked.assert_any_call(error_msg)

    def test_warns_if_settings_prefix_is_missing(self):
        config = Configurator()
        with mock.patch("kinto.core.initialization.warnings.warn") as mocked:
            kinto.core.initialize(config, "0.0.1")
            error_msg = "No value specified for `settings_prefix`"
            mocked.assert_any_call(error_msg)

    def test_set_the_settings_prefix_if_specified(self):
        config = Configurator()
        kinto.core.initialize(config, "0.0.1", "kinto")
        self.assertEqual(config.registry.settings["settings_prefix"], "kinto")

    def test_set_the_settings_prefix_from_settings_even_if_specified(self):
        config = Configurator(settings={"kinto.settings_prefix": "kinto"})
        kinto.core.initialize(config, "0.0.1", "readinglist")
        self.assertEqual(config.registry.settings["settings_prefix"], "kinto")

    def test_warns_if_not_https(self):
        with mock.patch("kinto.core.initialization.warnings.warn") as mocked:
            config = Configurator(
                settings={"settings_prefix": "kinto", "kinto.http_scheme": "https"}
            )
            kinto.core.initialize(config, "0.0.1")
            self.assertFalse(mocked.called)

            config = Configurator(settings={"kinto.http_scheme": "http"})
            kinto.core.initialize(config, "0.0.1")
            error_msg = "HTTPS is not enabled"
            mocked.assert_any_call(error_msg)

    def test_default_settings_are_overriden_if_specified_in_initialize(self):
        config = Configurator()
        defaults = {"paginate_by": 102}
        kinto.core.initialize(config, "0.0.1", "prefix", default_settings=defaults)
        self.assertEqual(config.registry.settings["paginate_by"], 102)

    def test_default_settings_are_overriden_by_application(self):
        config = Configurator(settings={"prefix.paginate_by": 10})
        kinto.core.initialize(config, "0.0.1", "prefix")
        self.assertEqual(config.registry.settings["paginate_by"], 10)

    def test_specified_default_settings_are_overriden_by_application(self):
        config = Configurator(settings={"settings_prefix.paginate_by": 5})
        defaults = {"settings_prefix.paginate_by": 10}
        kinto.core.initialize(config, "0.0.1", "settings_prefix", default_settings=defaults)
        self.assertEqual(config.registry.settings["paginate_by"], 5)

    def test_backends_are_not_instantiated_by_default(self):
        config = Configurator(settings={**kinto.core.DEFAULT_SETTINGS})
        kinto.core.initialize(config, "0.0.1", "settings_prefix")
        self.assertFalse(hasattr(config.registry, "storage"))
        self.assertFalse(hasattr(config.registry, "cache"))
        self.assertFalse(hasattr(config.registry, "permission"))

    def test_backends_type_is_checked_when_instantiated(self):
        def config_fails(settings):
            config = Configurator(settings=settings)
            with self.assertRaises(ConfigurationError):
                kinto.core.initialize(config, "0.0.1", "settings_prefix")

        config_fails({"settings_prefix.storage_backend": "kinto.core.cache.memory"})
        config_fails({"settings_prefix.cache_backend": "kinto.core.storage.memory"})
        config_fails({"settings_prefix.permission_backend": "kinto.core.storage.memory"})

    def test_environment_values_override_configuration(self):
        import os

        envkey = "KINTO_SETTINGS_PREFIX"
        os.environ[envkey] = "abc"

        config = Configurator(settings={"kinto.settings_prefix": "kinto"})
        kinto.core.initialize(config, "0.0.1")

        os.environ.pop(envkey)

        prefix_used = config.registry.settings["settings_prefix"]
        self.assertEqual(prefix_used, "abc")


class ProjectSettingsTest(unittest.TestCase):
    def settings(self, provided):
        config = Configurator(settings=provided)
        kinto.core.initialize(config, "1.0", "myproject")
        return config.get_settings()

    def test_uses_unprefixed_name(self):
        settings = {"paginate_by": 3.14}
        self.assertEqual(self.settings(settings)["paginate_by"], 3.14)

    def test_uses_settings_prefix(self):
        settings = {"myproject.paginate_by": 42}
        self.assertEqual(self.settings(settings)["paginate_by"], 42)

    def test_does_raise_valueerror_if_multiple_entries_are_equal(self):
        settings = {"paginate_by": 42, "myproject.paginate_by": 42}
        self.settings(settings)  # Not raising.

    def test_does_raise_valueerror_if_entries_are_not_hashable(self):
        settings = {"events.listeners": ["Rémy", 42]}
        self.settings(settings)  # Not raising.

    def test_raises_valueerror_if_different_multiple_entries(self):
        settings = {"paginate_by": 42, "myproject.paginate_by": 3.14}
        with self.assertRaises(ValueError):
            self.settings(settings)

    def test_environment_can_specify_settings_prefix(self):
        import os

        envkey = "MYPROJECT_CACHE_BACKEND"  # MYPROJECT_ prefix
        os.environ[envkey] = "kinto.core.cache.memory"
        settings = {"kinto.cache_backend": "kinto.core.cache.memcached"}
        value = self.settings(settings)["cache_backend"]
        os.environ.pop(envkey)
        self.assertEqual(value, "kinto.core.cache.memory")


class ApplicationWrapperTest(unittest.TestCase):
    @unittest.skipIf(initialization.newrelic is None, "newrelic is not installed.")
    @mock.patch("kinto.core.initialization.newrelic.agent")
    def test_newrelic_is_included_if_defined(self, mocked_newrelic):
        settings = {"newrelic_config": "/foo/bar.ini", "newrelic_env": "test"}
        mocked_newrelic.WSGIApplicationWrapper.return_value = "wrappedApp"
        app = kinto.core.install_middlewares(mock.sentinel.app, settings)
        mocked_newrelic.initialize.assert_called_with("/foo/bar.ini", "test")
        self.assertEqual(app, "wrappedApp")

    @unittest.skipIf(initialization.newrelic is None, "newrelic is not installed.")
    @mock.patch("kinto.core.initialization.newrelic.agent")
    def test_newrelic_is_not_included_if_set_to_false(self, mocked_newrelic):
        settings = {"newrelic_config": False}
        app = kinto.core.install_middlewares(mock.sentinel.app, settings)
        mocked_newrelic.initialize.assert_not_called()
        self.assertEqual(app, mock.sentinel.app)

    @mock.patch("kinto.core.initialization.ProfilerMiddleware")
    def test_profiler_is_not_installed_if_set_to_false(self, mocked_profiler):
        settings = {"profiler_enabled": False}
        app = kinto.core.install_middlewares(mock.sentinel.app, settings)
        mocked_profiler.initialize.assert_not_called()
        self.assertEqual(app, mock.sentinel.app)

    @mock.patch("kinto.core.initialization.ProfilerMiddleware")
    def test_profiler_is_installed_if_set_to_true(self, mocked_profiler):
        settings = {"profiler_enabled": True, "profiler_dir": "/tmp/path"}
        mocked_profiler.return_value = "wrappedApp"
        app = kinto.core.install_middlewares(mock.sentinel.app, settings)

        mocked_profiler.assert_called_with(
            mock.sentinel.app, restrictions="*kinto.core*", profile_dir="/tmp/path"
        )

        self.assertEqual(app, "wrappedApp")

    def test_load_default_settings_handle_prefix_attributes(self):
        config = mock.MagicMock()

        settings = {"settings_prefix": "myapp"}

        config.get_settings.return_value = settings
        initialization.load_default_settings(
            config,
            {
                "multiauth.policies": "basicauth",
                "myapp.http_scheme": "https",
                "myapp.http_host": "localhost:8888",
            },
        )

        self.assertIn("multiauth.policies", settings)
        self.assertNotIn("policies", settings)
        self.assertEqual(settings["multiauth.policies"], "basicauth")

        self.assertIn("http_scheme", settings)
        self.assertNotIn("kinto.http_scheme", settings)
        self.assertEqual(settings["http_scheme"], "https")

        self.assertIn("http_host", settings)
        self.assertNotIn("myapp.http_host", settings)
        self.assertEqual(settings["http_host"], "localhost:8888")

    def test_load_default_settings_converts_to_native_correctly(self):
        config = mock.MagicMock()

        settings = {"settings_prefix": "myapp"}

        config.get_settings.return_value = settings
        initialization.load_default_settings(config, {"myapp.my_cool_setting": '"1.2"'})

        self.assertEqual(settings["my_cool_setting"], "1.2")


class StatsDConfigurationTest(unittest.TestCase):
    def setUp(self):
        settings = {
            **kinto.core.DEFAULT_SETTINGS,
            "statsd_url": "udp://host:8080",
            "multiauth.policies": "basicauth",
        }
        self.config = Configurator(settings=settings)
        self.config.registry.storage = {}
        self.config.registry.cache = {}
        self.config.registry.permission = {}

        patch = mock.patch("kinto.core.statsd.load_from_config")
        self.mocked = patch.start()
        self.addCleanup(patch.stop)

    def test_statsd_isnt_called_if_statsd_url_is_not_set(self):
        self.config.add_settings({"statsd_url": None})
        initialization.setup_statsd(self.config)
        self.mocked.assert_not_called()

    def test_statsd_is_set_to_none_if_statsd_url_not_set(self):
        self.config.add_settings({"statsd_url": None})
        initialization.setup_statsd(self.config)
        self.assertEqual(self.config.registry.statsd, None)

    def test_statsd_is_called_if_statsd_url_is_set(self):
        initialization.setup_statsd(self.config)
        self.mocked.assert_called_with(self.config)
        # See `tests/core/test_statsd.py` for instantiation tests.

    def test_statsd_is_expose_in_the_registry_if_url_is_set(self):
        initialization.setup_statsd(self.config)
        self.assertEqual(self.config.registry.statsd, self.mocked.return_value)

    def test_statsd_is_set_on_cache(self):
        c = initialization.setup_statsd(self.config)
        c.watch_execution_time.assert_any_call({}, prefix="backend")

    def test_statsd_is_set_on_storage(self):
        c = initialization.setup_statsd(self.config)
        c.watch_execution_time.assert_any_call({}, prefix="backend")

    def test_statsd_is_set_on_permission(self):
        c = initialization.setup_statsd(self.config)
        c.watch_execution_time.assert_any_call({}, prefix="backend")

    def test_statsd_is_set_on_authentication(self):
        c = initialization.setup_statsd(self.config)
        c.watch_execution_time.assert_any_call(None, prefix="authentication")

    def test_statsd_counts_nothing_on_anonymous_requests(self):
        kinto.core.initialize(self.config, "0.0.1", "settings_prefix")
        app = webtest.TestApp(self.config.make_wsgi_app())
        app.get("/")
        self.assertFalse(self.mocked.count.called)

    def test_statsd_counts_views_and_methods(self):
        kinto.core.initialize(self.config, "0.0.1", "settings_prefix")
        app = webtest.TestApp(self.config.make_wsgi_app())
        app.get("/v0/__heartbeat__")
        self.mocked().count.assert_any_call("view.heartbeat.GET")

    def test_statsd_counts_unknown_urls(self):
        kinto.core.initialize(self.config, "0.0.1", "settings_prefix")
        app = webtest.TestApp(self.config.make_wsgi_app())
        app.get("/v0/coucou", status=404)
        self.assertFalse(self.mocked.count.called)

    @mock.patch("kinto.core.utils.hmac_digest")
    def test_statsd_counts_unique_users(self, digest_mocked):
        digest_mocked.return_value = "mat"
        kinto.core.initialize(self.config, "0.0.1", "settings_prefix")
        app = webtest.TestApp(self.config.make_wsgi_app())
        headers = {"Authorization": "Basic bWF0Og=="}
        app.get("/v0/", headers=headers)
        self.mocked().count.assert_any_call("users", unique="basicauth.mat")

    def test_statsd_counts_authentication_types(self):
        kinto.core.initialize(self.config, "0.0.1", "settings_prefix")
        app = webtest.TestApp(self.config.make_wsgi_app())
        headers = {"Authorization": "Basic bWF0Og=="}
        app.get("/v0/", headers=headers)
        self.mocked().count.assert_any_call("authn_type.basicauth")


class RequestsConfigurationTest(unittest.TestCase):
    def _get_app(self, settings={}):
        app_settings = {
            "storage_backend": "kinto.core.storage.memory",
            "cache_backend": "kinto.core.cache.memory",
        }
        app_settings.update(**settings)
        config = Configurator(settings=app_settings)
        kinto.core.initialize(config, "0.0.1", "name")
        return webtest.TestApp(config.make_wsgi_app())

    def test_requests_have_a_bound_data_attribute(self):
        config = Configurator()
        kinto.core.initialize(config, "0.0.1", "name")

        def on_new_request(event):
            data = event.request.bound_data
            self.assertEqual(data, {})
            self.assertEqual(id(data), id(event.request.bound_data))

        config.add_subscriber(on_new_request, NewRequest)
        app = webtest.TestApp(config.make_wsgi_app())
        app.get("/v0/")

    def test_subrequests_share_parent_bound_data(self):
        config = Configurator()
        kinto.core.initialize(config, "0.0.1", "name")

        bound_datas = set()

        def on_new_request(event):
            bound_datas.add(id(event.request.bound_data))

        config.add_subscriber(on_new_request, NewRequest)
        app = webtest.TestApp(config.make_wsgi_app())
        app.post_json("/v0/batch", {"requests": [{"path": "/"}]})
        self.assertEqual(len(bound_datas), 1)

    def test_by_default_relies_on_pyramid_application_url(self):
        app = self._get_app()
        resp = app.get("/v0/")
        self.assertEqual(resp.json["url"], "http://localhost/v0/")

    def test_by_default_relies_on_incoming_headers(self):
        app = self._get_app()
        resp = app.get("/v0/", headers={"Host": "server:8888"})
        self.assertEqual(resp.json["url"], "http://server:8888/v0/")

    def test_by_default_relies_on_wsgi_environment(self):
        app = self._get_app()
        environ = {"wsgi.url_scheme": "https", "HTTP_HOST": "server:44311"}
        resp = app.get("/v0/", extra_environ=environ)
        self.assertEqual(resp.json["url"], "https://server:44311/v0/")

    def test_http_scheme_overrides_the_wsgi_environment(self):
        app = self._get_app({"http_scheme": "http2"})
        environ = {"wsgi.url_scheme": "https"}
        resp = app.get("/v0/", extra_environ=environ)
        self.assertEqual(resp.json["url"], "http2://localhost:80/v0/")

    def test_http_host_overrides_the_wsgi_environment(self):
        app = self._get_app({"http_host": "server"})
        environ = {"HTTP_HOST": "elb:44311"}
        resp = app.get("/v0/", extra_environ=environ)
        self.assertEqual(resp.json["url"], "http://server/v0/")

    def test_http_host_overrides_the_request_headers(self):
        app = self._get_app({"http_host": "server"})
        resp = app.get("/v0/", headers={"Host": "elb:8888"})
        self.assertEqual(resp.json["url"], "http://server/v0/")


class PluginsTest(unittest.TestCase):
    def test_kinto_core_includes_are_included_manually(self):
        config = Configurator(settings={**kinto.core.DEFAULT_SETTINGS})
        config.add_settings({"includes": "elastic history"})
        config.route_prefix = "v2"

        with mock.patch.object(config, "include"):
            with mock.patch.object(config, "scan"):
                kinto.core.includeme(config)

                config.include.assert_any_call("elastic")
                config.include.assert_any_call("history")

    def make_app(self):
        config = Configurator(settings={**kinto.core.DEFAULT_SETTINGS})
        config.add_settings(
            {
                "permission_backend": "kinto.core.permission.memory",
                "includes": "tests.core.testplugin",
                "multiauth.policies": "basicauth",
            }
        )
        kinto.core.initialize(config, "0.0.1", "name")
        return webtest.TestApp(config.make_wsgi_app())

    def test_plugin_can_define_protected_views(self):
        app = self.make_app()
        app.post("/v0/attachment", status=401)
        headers = {"Authorization": "Basic bWF0OjE="}
        app.post("/v0/attachment", headers=headers, status=403)

    def test_plugin_benefits_from_cors_setup(self):
        app = self.make_app()
        headers = {"Origin": "lolnet.org", "Access-Control-Request-Method": "POST"}
        resp = app.options("/v0/attachment", headers=headers, status=200)
        self.assertIn("Access-Control-Allow-Origin", resp.headers)



import logging

import transaction as current_transaction
from pyramid.settings import asbool

from kinto.plugins.quotas import scripts as quotas

logger = logging.getLogger(__name__)


def rebuild_quotas(env, dry_run=False):
    """Administrative command to rebuild quota usage information.

    This command recomputes the amount of space used by all
    collections and all buckets and updates the quota objects in the
    storage backend to their correct values. This can be useful when
    cleaning up after a bug like e.g.
    https://github.com/Kinto/kinto/issues/1226.
    """
    registry = env["registry"]
    settings = registry.settings
    readonly_mode = asbool(settings.get("readonly", False))

    # FIXME: readonly_mode is not meant to be a "maintenance mode" but
    # rather used with a database user that has read-only permissions.
    # If we ever introduce a maintenance mode, we should maybe enforce
    # it here.
    if readonly_mode:
        message = "Cannot rebuild quotas while in readonly mode."
        logger.error(message)
        return 41

    if "kinto.plugins.quotas" not in settings["includes"]:
        message = "Cannot rebuild quotas when quotas plugin is not installed."
        logger.error(message)
        return 42

    quotas.rebuild_quotas(registry.storage, dry_run=dry_run)
    current_transaction.commit()
    return 0
